
import skparser
import indenter
import sys

# convenience declarations for types/objects that may be used...

TV_NONE    = skparser.ExprAttr.TV_NONE
TV_BOOLLIT = skparser.ExprAttr.TV_BOOLLIT
TV_INTLIT  = skparser.ExprAttr.TV_INTLIT
TV_REALLIT = skparser.ExprAttr.TV_REALLIT
TV_STRLIT  = skparser.ExprAttr.TV_STRLIT
TV_BOOL    = skparser.ExprAttr.TV_BOOL
TV_INTEGER = skparser.ExprAttr.TV_INTEGER
TV_REAL    = skparser.ExprAttr.TV_REAL
TV_STRING  = skparser.ExprAttr.TV_STRING
TV_OBJECT  = skparser.ExprAttr.TV_OBJECT

QT_NONE    = skparser.SymbolAttr.QT_NONE
QT_PARAM   = skparser.SymbolAttr.QT_PARAM
QT_LOCAL   = skparser.SymbolAttr.QT_LOCAL
QT_GLOBAL  = skparser.SymbolAttr.QT_GLOBAL

prbuff = indenter.prbuff

###########################################################################
#
# type help
#
# note:  10/13 - doesn't seem to work the same from command line as does in envir
# so skip for now.
#

def islist(x):
    comp = []
    if x.__class__ == comp.__class__:
        return True
    return False

def isstring(x):
    if type(x) == type(""):
        return True
    return False

def istuple(x):
    if type(x) == type(()):
        return True;
    return False;

def isnone(x):
    if type(x) == type(None):
        return True;
    return False

def isfunc(x):
    def _f_(): pass
    if type(x) == type(_f_):
        return True
    return False

def quoted(x):
    if isstring(x):
        if x[0] == "\'" or x[0] == "\"":
            return x
        else:
            return "\"" + x + "\""
    # just let it fail for now
    return ""

###########################################################################
#
# base class for rw-writer objects.  there are four primary methods to be
# overloaded (all of these are static methods):
#
#   @staticmethod
#   who():
#     -> id of the function that object rewrites
#
#   @staticmethod
#   type():
#     -> the return type of the function call.
#
#   @staticmethod
#   strrep(args):
#     -> return a string representation of the function call
#
#   @staticmethod
#   pp(args):
#     -> call the pretty-printer interface to representat the function call
#     (default is to call through to strrep).
#
# refer to the following for example implementations.
#
# re-writer objects must be registered with the frontend.
#
# NOTES:
#   -- TODO:  put argcnt check in methods...

class RWIntfc:
    def __init__(self):
        pass

    @staticmethod
    def who():
        return ""

    @staticmethod
    def type():
        return TV_NONE

    @staticmethod
    def strargs(args):
        # utility function to just generate string rep of args
        ret = "("
        for arg in args:
            ret += arg.strrep() + ", "
        # trim the last comma-separator (if there were arguments)
        if len(args) > 0:
            ret = ret[:-2]
        ret += ")"
        return ret

    @staticmethod
    def ppargs(redir, args):
        # utility function to prettyprint args
        prbuff.substr("(")
        argcnt = len(args)
        for idx in range(0, argcnt):
            args[idx].pppython(redir)
            if idx < argcnt - 1:
                prbuff.substr(",")
                prbuff.blank(1)
        prbuff.substr(")")

    @staticmethod
    def strrep(args):
        return ""

    @staticmethod
    def pp(redir, args):
        # default behavior is to just do the string representation
        prbuff.substr(RWIntfc.strrep(args))

#
#  cell_id = dbOpenCellViewByType( librarry_name cell_name view_name view_type open_mode)
#    library_name - the library that the cell is located in
#    cell_name - the name of the cell that is being opened
#    view_name - name of view being opened (layout, symbol)
#    view_type - type of the view being opened (maskLayout) (opt)
#    open_mode - how it is opened ("w" - write/overwrite, "a" - append, "r" - read) (opt)
#

class DbOpenCellView(RWIntfc):
    def __init__(self):
        RWIntfc.__init__(self)

    @staticmethod
    def who():
        return "dbOpenCellView"

    @staticmethod
    def type():
        return TV_OBJECT

    @staticmethod
    def strrep(args):
        rep = "self"
        return rep

    @staticmethod
    def pp(redir, args):
        prbuff.substr("self")

#  dbClose(cell_id)
#    Used to close an open cellview - doesn't save.  (not supported -
#  so just ignore).
#

class DbClose(RWIntfc):
    def __init__(self):
        RWIntfc.__init__(self)

    @staticmethod
    def who():
        return "dbClose"

    @staticmethod
    def type():
        return TV_OBJECT

    @staticmethod
    def strrep(args):
        rep = ""
        
    @staticmethod
    def pp(redir, args):
        prbuff.substr("")

#  dbSave( cell_id )
#    Used to save the changes (not supported - so just ignore)
#

class DbSave(RWIntfc):
    def __init__(self):
        RWIntfc.__init__(self)

    @staticmethod
    def who():
        return "dbSave"

    @staticmethod
    def type():
        return TV_OBJECT

    @staticmethod
    def strrep(args):
        rep = ""

    @staticmethod
    def pp(redir, args):
        prbuff.substr("")

#  dbCreateParamInst( work_cell_id cell_id inst_name location rotation number params) => inst_id
#    Used to place an parameterized instance into the work_cell_view
#
#    work_cell_id - the cellview where instance is placed
#    cell_id - the instance that you want to add (must use dbOpenCellViewByType())
#    inst_name - name the instance if nil (program assigns unique name)
#    location (list) - x:y position
#    rotation - orientation of instance ("R0","R90","R180","R270",...)
#    number - number of instances to place at location (opt)
#    params - list of parameters that you want to set explicitly others are
#    defaulted to orignal setting
#
#    For PyCells, the Instance class is the base class for all DLO objects.    The
#    constructor is as follows:
#
#      Instance(string dloName, params=None, string nodelist="", string instname="")
#
#    dloName - string of the from "libName/cellName/viewName"
#      pars = ParamArray()
#      inst = Instance("libname/cellname/viewname", pars, 'd g s b', 'inst1')
#

class DbCreateInst(RWIntfc):
    def __init__(self):
        RWIntfc.__init__(self)

    @staticmethod
    def who():
        return "dbCreateInst"

    @staticmethod
    def type():
        return TV_OBJECT

    @staticmethod
    def strrep(args):
        str = "Instance("
        argcnt = len(args)
        if argcnt != 6:
          str += ")"
          return str

        dloname = args[1].strrep()
        # following is not correct, but need to use for now
        newargs = [args[3], args[4], args[2]]
        rep = "Instance" + RWIntfc.strargs(newargs)
        return rep

    @staticmethod
    def pp(redir, args):
        # call base for now...
        RWIntfc.pp(redir, args)

class DbCreateParamInst(RWIntfc):
    def __init__(self):
        RWIntfc.__init__(self)

    @staticmethod
    def who():
        return "dbCreateParamInst"

    @staticmethod
    def type():
        return TV_OBJECT

    @staticmethod
    def strrep(args):
        str = "Instance("
        argcnt = len(args)
        if argcnt != 7:        # should be 7 args for parameterized instances
          str += ")"
          return str

        dloname = args[1].strrep()
        # following is not correct, but need to use for now
        newargs = [args[3], args[4], args[6], args[2]]
        rep = "Instance" + RWIntfc.strargs(newargs)
        return rep

    @staticmethod
    def pp(redir, args):
        # call base for now...
        RWIntfc.pp(redir, args)
        

#  dbMakeNet( work_cell_view net_name) => net_id
#  dbCreateNet( work_cell_view net_name) => ned_id
#
#    Used to create a net of net_name
#
#    net_name - name of the net to create
#
#    for PyCells, the Net class represents connectivity in a design.
#
#      net = Net('name')
#

class DbMakeNet(RWIntfc):
    def __init__(self):
        RWIntfc.__init__(self)

    @staticmethod
    def who():
        return "dbMakeNet"

    @staticmethod
    def type():
        return TV_OBJECT

    @staticmethod
    def strrep(args):
        rep = "Net("
        if len(args) < 2:
            rep += ")"
        else:
            rep += args[1].strrep() + ")"
        return rep

    @staticmethod
    def pp(redir, args):
        prbuff.substr("Net(")
        if len(args) < 2:
            prbuff.substr(")")
        else:
            args[1].ppython(redir)
            prbuff.substr(")")
        return

class DbCreateNet(RWIntfc):
    def __init__(self):
        RWIntfc.__init__(self)

    @staticmethod
    def who():
        return "dbCreateNet"

    @staticmethod
    def type():
        return TV_OBJECT

    @staticmethod
    def strrep(args):
        rep = "Net("
        if len(args) < 2:
            rep += ")"
        else:
            rep += args[1].strrep() + ")"
        return rep

    @staticmethod
    def pp(redir, args):
        prbuff.substr("Net(")
        if len(args) < 2:
            prbuff.substr(")")
        else:
            args[1].ppython(redir)
            prbuff.substr(")")
        return
        
#  dbCreateTerm(net_id term_name term_dir) => term_id
#    Used to create a terminal for a net.
#
#    net_id - name of the net
#    term_name - name of the terminal (if nil then term_name = term_id)
#    term_dir - direction of the terminal ("input","output","inputOutput")
#
#    eg.  trm = dbCreateTerm(net "ip1" "input")
#

class DbCreateTerm(RWIntfc):
    def __init__(self):
        RWIntfc.__init__(self)

    @staticmethod
    def who():
        return "dbCreateTerm"

    @staticmethod
    def type():
        return TV_OBJECT

#
# api not as advertised - don't pass direction on
#   @staticmethod
#   def strrep(args):
#       rep = "self.addTerm"
#       if len(args) < 2:
#           rep += "()"
#           return rep
#       netname = args[0].strrep() + ".name"
#       dir = args[1].strrep()
#       if len(args) == 3:
#           dir = args[2].strrep()
#       rep += "(" + netname + ", " + dir + ")"
#       return rep

    @staticmethod
    def strrep(args):
        rep = "self.addTerm"
        if len(args) < 2:
            rep += "()"
            return rep
        netname = args[0].strrep() + ".name"
        rep += "(" + netname + ")"
        return rep

    @staticmethod
    def pp(redir, args):
        prbuff.substr("self.addTerm")
        if len(args) < 2:
            prbuff.substr("()")
            return
        netname = args[0].strrep() + ".name"
        dir = args[1].strrep()
        if len(args) == 3:
            dir = args[2].strrep()
        prbuff.substr("(")
        prbuff.substr(netname)
        prbuff.substr(", ")
        prbuff.substr(dir)
        prbuff.substr(")")
        return rep
    
#  dbCreatePin(net_id shape pin_name) => pin
#    Used to create a pin for a net with a figure shape
#
#    net_id - name of the net
#    shape - figure of the pin (created using dbCreateRect, dbCreatePolygon, etc...)
#    pin_name - name of the pin (if nil then name automatically generated) (optional)
#
#    for PyCells, whenever a Contact is created, the system will automatically
#    create a pin object.
#
#      contact1 = Contact(Layer('metal1'), Layer('metal3'), 'A')
#      pin = contact1.firstPin()
#
# notes:  pins describe physical connections and are associated to a specific
# terminal (a terminal may have more than one pin).  a pin is also associated
# with a physical shape in the design.
#

class DbCreatePin(RWIntfc):
    def __init__(self):
        RWIntfc.__init__(self)

    @staticmethod
    def who():
        return "dbCreatePin"

    @staticmethod
    def type():
        return TV_OBJECT

# api kinda weak - taking a rect figure should be same as
# specifying a bounding box and a layer.
#   @staticmethod
#   def strrep(args):
#       rep = "self.addPin"
#       argcnt = len(args)
#       if argcnt < 2:
#           rep += "()"
#           return rep
#       termname = args[0].strrep() + ".name"
#       pinname = termname
#       if argcnt == 3:
#          pinname = args[2].strrep()
#       region = args[1].strrep()
#       rep += "(" + pinname + ", " + termname + ", " + region + ")"
#       return rep

    @staticmethod
    def strrep(args):
        rep = "self.addPin"
        argcnt = len(args)
        if argcnt < 2:
            rep += "()"
            return rep
        termname = args[0].strrep() + ".name"
        pinname = termname
        if argcnt == 3:
           pinname = args[2].strrep()
        region = args[1].strrep()
        box = region + ".getBBox()"
        layer = region + ".getLayer()"
        rep += "(" + pinname + ", " + termname + ", " + box + ", " + layer + ")"
        return rep

    @staticmethod
    def pp(redir, args):
        prbuff.substr("self.addPin")
        argcnt = len(args)
        if argcnt < 2:
            prbuff.substr("()")
            return
        termname = args[0].strrep() + ".name"
        pinname = termname
        if argcnt == 3:
           pinname = args[2].strrep()
        region = args[1].strrep()
        prbuff.substr("(")
        prbuff.substr(pinname)
        prbuff.substr(",")
        prbuff.blank(1)
        prbuff.substr(termname)
        prbuff.substr(",")
        prbuff.blank(1)
        prbuff.substr(region)
        prbuff.substr(")")

#  dbCreateRect(work_cell_id layer_name box) => shape
#    Used to create a rectangle in layer layer_name
#
#    work_cell_id - cellview that adding the rectangle to
#    layer_name - name of layer making rectangle in
#    box - list of points indicating points of rectangle
#
#    fig = dbCreateRect(cv list("m1" "drawing") list(x:y x+5:y+5))
#
#    For Pycells, rects are defined by bounding boxes.
#      rect = Rect(Layer('layer1'), Box(left, bottom, right, top))
#

class DbCreateRect(RWIntfc):
    def __init__(self):
        RWIntfc.__init__(self)

    @staticmethod
    def who():
        return "dbCreateRect"

    @staticmethod
    def type():
        return TV_OBJECT

    @staticmethod
    def strrep(args):
        rep = "Rect("
        argcnt = len(args)
        if argcnt != 3:
           str += ")"
           return str
        layer = args[1].strrep()
        #if islist(layer):
            #layer = layer[0].strrep()
        #layer = quoted(layer)
        if args[1].is_container():
            layer = args[1][0].strrep()
        layer = quoted(layer)
        rep += "Layer(" + layer + "), Box("
        list = args[2]
        ptcnt = len(list)
        if ptcnt != 2:
            rep += "))"
            return rep
        rep += list[0].strrep()
        rep += ", "
        rep += list[1].strrep()
        rep +=  "))"
        return rep

    @staticmethod
    def pp(redir, args):
        prbuff.substr("Rect(")
        argcnt = len(args)
        if argcnt != 3:
           str += ")"
           prbuff.substr(str)
           return

        prbuff.substr("Layer(")
        prbuff.blank(0)
        args[1].pppython(redir)
        prbuff.substr("),")
        prbuff.blank(1)
        args[2].pppython(redir)
        prbuff.blank(0)
        prbuff.substr(")")


#  dbCreatePolygon(work_cell_id layer_name poly) => shape
#    Used to create a polygon in layer layer_name that
#    automatically closes it self
#
#    work_cell_id - cellview that adding the rectangle to
#    layer_name - name of layer making rectangle in
#    poly - list of points indicating points of polygon first point is where
#    last point will connect to
#
#    eg. fig = dbCreateRect(cv list("m1" "drawing") list(x:y x+5:y x+5:y+5 x:y+5))
#
#    In PyCells, this woould be represented:
#
#      poly = Polygon(Layer('layer1'), [Point(x1,y1), Point(x2,y2), ... Point(xn,yn)])
#

class DbCreatePolygon(RWIntfc):
    def __init__(self):
        RWIntfc.__init__(self)

    @staticmethod
    def who():
        return "dbCreatePolygon"

    @staticmethod
    def type():
        return TV_OBJECT

    @staticmethod
    def strrep(args):
        rep = "Polygon("
        argcnt = len(args)
        if argcnt != 4:
           rep += ")"
           return rep
        layer = args[1].strrep()
        if args[1].is_container():
            layer = args[1][0].strrep()
        layer = quoted(layer)
        rep += "Layer(" + layer + "),"

        # wrap up the list of points...
        list = args[2]
        argcnt = len(list.elist)
        rep += "points=["
        for idx in range(0, argcnt):
            stmp = list[idx].strrep()
            rep += stmp
            if i < argcnt - 1:
                rep += ", "

        rep += "])"
        return rep

    @staticmethod
    def pp(redir, args):
        prbuff.substr("Polygon(")
        argcnt = len(args)
        if argcnt != 4:
           str += ")"
           prbuff.substr(str)
           return

        prbuff.substr("Layer(")
        prbuff.blank(0)
        args[1].pppython(redir)
        prbuff.substr("),")
        
        # wrap up the list of points...
        ptlist = args[2]
        ptcnt = len(ptlist.elist)
        prbuff.substr("points=[")
        for idx in range(0, ptcnt):
            ptlist[idx].pppython(redir)
            if i < ptcnt - 1:
                prbuff.substr(",")
                prbuff.blank(1)
                
        prbuff.substr("]")
        prbuff.blank(0)
        prbuff.substr(")")

#  dbCreatePath(work_cell_id layer_name path width path_style) => shape
#    Used to create a rectangle in layer layer_name
#
#    work_cell_id - cellview that adding the rectangle to
#    layer_name - name of layer making rectangle in
#    path - list of points indicating points of path
#    width - width of the path (um)
#    path_style - type of path ("truncateExtend","offset","squareFlush","extendExtend") (optional)

class DbCreatePath(RWIntfc):
    def __init__(self):
        RWIntfc.__init__(self)

    @staticmethod
    def who():
        return "dbCreatePath"

    @staticmethod
    def type():
        return TV_OBJECT

    @staticmethod
    def strrep(args):
        rep = "Path("
        argcnt = len(args)
        if argcnt != 4:
           rep += ")"
           return rep

        layer = args[1].strrep()
        if args[1].is_container():
            layer = args[1][0].strrep()
        layer = quoted(layer)
        rep += "Layer(" + layer + "),"

        # wrap up the list of points...
        list = args[2]
        argcnt = len(list.elist)
        rep += "points=["
        for idx in range(0, argcnt):
            stmp = list[idx].strrep()
            rep += stmp
            if i < argcnt - 1:
                rep += ", "

        rep += "])"
        return rep

    @staticmethod
    def pp(redir, args):
        prbuff.substr("Path(")
        argcnt = len(args)
        if argcnt != 4:
           prbuff.substr(")")
           return

        prbuff.substr("Layer(")
        prbuff.blank(0)
        args[1].pppython(redir)
        prbuff.substr("),")

        # wrap up the list of points...
        ptlist = args[2]
        ptcnt = len(ptlist.elist)
        prbuff.substr("points=[")
        for idx in range(0, ptcnt):
            ptlist[idx].pppython(redir)
            if i < ptcnt - 1:
                prbuff.substr(",")
                prbuff.blank(1)

        prbuff.substr("]")
        prbuff.blank(0)
        prbuff.substr(")")

#########################################################################


class DdGetObj(RWIntfc):
    def __init__(self):
        RWIntfc.__init__(self)

    @staticmethod
    def who():
        return "ddGetObj"

    @staticmethod
    def type():
        return TV_OBJECT

    @staticmethod
    def strrep(args):
        rep = "ddGetObj" + RWIntfc.strargs(args)
        return rep

    @staticmethod
    def pp(redir, args):
        prbuff.substr("ddGetObj")
        RWIntfc.ppargs(redir, args)


class MathFloor(RWIntfc):
    def __init__(self):
        RWIntfc.__init__(self)

    @staticmethod
    def who():
        return "floor"

    @staticmethod
    def type():
        return TV_INTEGER

    @staticmethod
    def strrep(args):
        rep = "math.floor" + RWIntfc.strargs(args)
        return rep

    @staticmethod
    def pp(redir, args):
        prbuff.substr("math.floor")
        RWIntfc.ppargs(redir, args)


class MathCeiling(RWIntfc):
    def __init__(self):
        RWIntfc.__init__(self)

    @staticmethod
    def who():
        return "ceiling"

    @staticmethod
    def type():
        return TV_INTEGER

    @staticmethod
    def strrep(args):
        rep = "math.ceil" + RWIntfc.strargs(args)
        return rep

    @staticmethod
    def pp(redir, args):
        prbuff.substr("math.ceil")
        RWIntfc.ppargs(redir, args)


class Strcat(RWIntfc):
    def __init__(self):
        RWIntfc.__init__(self)

    @staticmethod
    def who():
        return "strcat"

    @staticmethod
    def type():
        return TV_STRING

    @staticmethod
    def strrep(args):
# in python - strcat maps to:  arg1 + arg2 + .. + argn
        rep = ""
        argcnt = len(args)
        for i in range(argcnt):
            rep += args[i].strrep()
            if i < argcnt - 1:
                rep += " + "
        return rep

    @staticmethod
    def pp(redir, args):
        argcnt = len(args)
        for i in range(argcnt):
            prbuff.substr(args[i].pppython(redir))
            if i < argcnt - 1:
                prbuff.blank(1)
                prbuff.substr("+")
                prbuff.blank(1)

class Sprintf(RWIntfc):
    def __init__(self):
        RWIntfc.__init__(self)

    @staticmethod
    def who():
        return "sprintf"

    @staticmethod
    def type():
        return TV_STRING

    @staticmethod
    def strrep(args):
# in python - sprintf maps to:  "arg2"%(arg3,arg4,...)
        rep = ""
        argcnt = len(args)
        if argcnt < 3:
            return rep
        rep = args[1].strrep()
        rep += "%("
        for i in range(2, argcnt):
            rep += args[i].strrep()
            if i < argcnt - 2:
                rep += ", "
        rep += ")"
        return rep

    @staticmethod
    def pp(redir, args):
        argcnt = len(args)
        if argcnt < 3:
            prbuff.substr("sprintf")
            return
        args[1].pppython(redir)
        prbuff.substr("(")
        for i in range(2, argcnt):
            prbuff.substr(args[i].pppython(redir))
            if i < argcnt - 2:
                prbuff.substr(",")
                prbuff.blank(1)
        prbuff.substr(")")

class Modulo(RWIntfc):
    def __init__(self):
        RWIntfc.__init__(self)

    @staticmethod
    def who():
        return "modulo"

    @staticmethod
    def type():
        return TV_INTEGER

    @staticmethod
    def strrep(args):
# in python - strcat maps to:  arg1 + arg2 + .. + argn
        rep = ""
        argcnt = len(args)
        if argcnt < 2:
            return rep
        rep = args[0].strrep()
        rep += "%"
        rep += args[1].strrep()
        return rep

    @staticmethod
    def pp(redir, args):
        argcnt = len(args)
        if argcnt < 2:
            prbuff.substr("insufficient number args for modulo")
            return
        args[0].pppython(redir)
        prbuff.substr("%")
        args[1].pppython(redir)

class Index(RWIntfc):
    def __init__(self):
        RWIntfc.__init__(self)

    @staticmethod
    def who():
        return "index"

    @staticmethod
    def type():
        return TV_STRING

    @staticmethod
    def strrep(args):
# in python - index maps to:  s1[s1.index(s2):]
        rep = ""
        argcnt = len(args)
        if argcnt != 2:
            return rep
        arg1 = args[0].strrep()
        arg2 = args[1].strrep()
        rep = arg1 + "[" + arg1 + ".index(" + arg2 + "):]"
        return rep

    @staticmethod
    def pp(redir, args):
        argcnt = len(args)
        if argcnt != 2:
            prbuff.substr("error with index()")
            return
        rep = ""
        arg1 = args[0].strrep()
        arg2 = args[1].strrep()
        prbuff.substr(arg1)
        prbuff.substr("[")
        prbuff.substr(arg1)
        prbuff.substr(".index(")
        prbuff.substr(arg2)
        prbuff.substr("):]")

class Rindex(RWIntfc):
    def __init__(self):
        RWIntfc.__init__(self)

    @staticmethod
    def who():
        return "rindex"

    @staticmethod
    def type():
        return TV_STRING

    @staticmethod
    def strrep(args):
# in python - rindex maps to:  s1[s1.rindex(s2):]
        rep = ""
        argcnt = len(args)
        if argcnt != 2:
            return rep
        arg1 = args[0].strrep()
        arg2 = args[1].strrep()
        rep = arg1 + "[" + arg1 + ".rindex(" + arg2 + "):]"
        return rep

    @staticmethod
    def pp(redir, args):
        argcnt = len(args)
        if argcnt != 2:
            prbuff.substr("error with rindex()")
            return
        rep = ""
        arg1 = args[0].strrep()
        arg2 = args[1].strrep()
        prbuff.substr(arg1)
        prbuff.substr("[")
        prbuff.substr(arg1)
        prbuff.substr(".rindex(")
        prbuff.substr(arg2)
        prbuff.substr("):]")

#########################################################################

#
# dbGetq is passed two arguments - the object and the attribute.
#

class DbGetq(RWIntfc):
    def __init__(self):
        RWIntfc.__init__(self)

    @staticmethod
    def who():
        return "dbGetq"

    @staticmethod
    def type():
        return TV_OBJECT

    @staticmethod
    def strrep(args):
        rep = "dbGetq"
        argcnt = len(args)
        if argcnt != 2:
            rep += ")"
            return rep
        rep += RWIntfc.strargs(args)
        return rep

    @staticmethod
    def pp(redir, args):
        prbuff.substr("dbGetq")
        argcnt = len(args)
        if argcnt != 2:
            prbuff.substr(")")
            return
        RWIntfc.ppargs(redir, args)


#
# dbSetq is passed three arguments - the object, attribute, and value
#

class DbSetq(RWIntfc):
    def __init__(self):
        RWIntfc.__init__(self)

    @staticmethod
    def who():
        return "dbSetq"

    @staticmethod
    def type():
        return TV_OBJECT

    @staticmethod
    def strrep(args):
        rep = "dbSetq"
        argcnt = len(args)
        if argcnt != 3:
            rep += ")"
            return rep
        rep += RWIntfc.strargs(args)
        return rep

    @staticmethod
    def pp(redir, args):
        prbuff.substr("dbSetq")
        argcnt = len(args)
        if argcnt != 3:
            prbuff.substr(")")
            return
        RWIntfc.ppargs(redir, args)


###########################################################################

#
# techfile rewriter functions
#

class TechGetTechFile(RWIntfc):
    def __init__(self):
        RWIntfc.__init__(self)

    @staticmethod
    def who():
        return "techGetTechFile"

    @staticmethod
    def type():
        return TV_OBJECT

    @staticmethod
    def strrep(args):
        # rep = "Tech.get(" + args[0].strrep() + ")"
        return "self.tech"

    @staticmethod
    def pp(redir, args):
        prbuff.substr("Tech.get")
        RWIntfc.ppargs(redir, args)

class TechGetMfgGridResolution(RWIntfc):
    def __init__(self):
        RWIntfc.__init__(self)

    @staticmethod
    def who():
        return "techGetMfgGridResolution"

    @staticmethod
    def type():
        return TV_REAL

    @staticmethod
    def strrep(args):
        return args[0].strrep() + ".getGridResolution()"

    @staticmethod
    def pp(redir, args):
        prbuff.substr(args[0].strrep())
        prbuff.substr(".getGridResolution()")

class TechGetSpacingRule(RWIntfc):
    def __init__(self):
        RWIntfc.__init__(self)

    @staticmethod
    def who():
        return "techGetSpacingRule"

    @staticmethod
    def type():
        return TV_OBJECT

    @staticmethod
    def strrep(args):
        rep = args[0].strrep() + ".getPhysicalRule(" + args[1].strrep()
        layer = args[2].strrep()
        if islist(layer):
            layer = layer[0].strrep()
        layer = quoted(layer)
        rep += ", Layer(" + layer + ")"
        if len(args) > 3:
            # additional arugment
            layer = args[3].strrep()
            if islist(layer):
                layer = layer[0].strrep()
            layer = quoted(layer)
            rep += ", Layer(" + layer + ")"
        rep += ")"
        return rep

    @staticmethod
    def pp(redir, args):
        prbuff.substr(args[0].strrep())
        prbuff.substr(".getPhysicalArgs")
        subargs = args[1:]
        RWIntfc.ppargs(redir, subargs)

class TechGetOrderedSpacingRule(RWIntfc):
    def __init__(self):
        RWIntfc.__init__(self)

    @staticmethod
    def who():
        return "techGetOrderedSpacingRule"

    @staticmethod
    def type():
        return TV_OBJECT

    @staticmethod
    def strrep(args):
        rep = args[0].strrep() + ".getPhysicalRule(" + args[1].strrep()
        rep += ", " + "Layer(" + args[2].strrep() + "), Layer(" + args[3].strrep() + "))"
        return rep

    @staticmethod
    def pp(redir, args):
        prbuff.substr(args[0].strrep())
        prbuff.substr(".getPhysicalArgs")
        subargs = args[1:]
        RWIntfc.ppargs(redir, subargs)

#
# there is no real corresponding ciranova API method for techGetParam - just do 
# the best - which probably means commenting out of the skill code and working
# around.
#
class TechGetParam(RWIntfc):
    def __init__(self):
        RWIntfc.__init__(self)

    @staticmethod
    def who():
        return "techGetParam"

    @staticmethod
    def type():
        return TV_OBJECT

    @staticmethod
    def strrep(args):
        rep = args[0].strrep() + ".techGetParam"
        largs = args[1:]
        rep += RWIntfc.strargs(largs)
        return rep

    @staticmethod
    def pp(redir, args):
        args[0].pppython(redir)
        prbuff.substr(".techGetParam")
        largs = args[1:]
        RWIntfc.ppargs(redir, largs)

#
# this is not in the SKILL API - but the data is typically used, and allows
# us to modify the skill code to get the translater to do the rest of the
# work.
#
class TechGetMosfetParams(RWIntfc):
    def __init__(self):
        RWIntfc.__init__(self)

    @staticmethod
    def who():
        return "techGetMosfetParams"

    @staticmethod
    def type():
        return TV_OBJECT

    @staticmethod
    def strrep(args):
        rep = args[0].strrep() + ".getMosfetParams"
        largs = args[1:]
        rep += RWIntfc.strargs(largs)
        return rep

    @staticmethod
    def pp(redir, args):
        args[0].pppython(redir)
        prbuff.substr(".getMosfetParams")
        largs = args[1:]
        RWIntfc.ppargs(redir, largs)

##########################################################################



class DbReplaceProp(RWIntfc):
    def __init__(self):
        RWIntfc.__init__(self)

    @staticmethod
    def who():
        return "dbReplaceProp"

    @staticmethod
    def type():
        return TV_OBJECT

    @staticmethod
    def strrep(args):
        rep = "# dbReplaceProp()"
#       argcnt = len(args)
#       if argcnt != 3:
#               rep += ")"
#               return rep
#       rep += args[0].strrep()
#       rep += ", "
#       rep += args[2].strrep()
#       rep += ")
        return rep

    @staticmethod
    def pp(redir, args):
        prbuff.substr("#dbReplaceProp()")


##########################################################################

def initRWIntfcObjs(frontend):
        # format is:  frontend.regIntfcObj(<rewriter object>)
        #
        frontend.regIntfcObj(DdGetObj)
        frontend.regIntfcObj(Strcat)
        frontend.regIntfcObj(MathFloor)
        frontend.regIntfcObj(MathCeiling)
        frontend.regIntfcObj(Sprintf)
        frontend.regIntfcObj(Modulo)
        frontend.regIntfcObj(Index)
        frontend.regIntfcObj(Rindex)
        frontend.regIntfcObj(DbOpenCellView)
        frontend.regIntfcObj(DbCreateInst)
        frontend.regIntfcObj(DbCreateParamInst)
        frontend.regIntfcObj(DbMakeNet)
        frontend.regIntfcObj(DbCreateTerm)
        frontend.regIntfcObj(DbCreatePin)
        frontend.regIntfcObj(DbCreateNet)
        frontend.regIntfcObj(DbCreateRect)
        frontend.regIntfcObj(DbCreatePolygon)
        frontend.regIntfcObj(DbCreatePath)
        frontend.regIntfcObj(DbSetq)
        frontend.regIntfcObj(DbGetq)
        frontend.regIntfcObj(DbReplaceProp)
        frontend.regIntfcObj(TechGetTechFile) 
        frontend.regIntfcObj(TechGetMfgGridResolution) 
        frontend.regIntfcObj(TechGetSpacingRule) 
        frontend.regIntfcObj(TechGetOrderedSpacingRule) 
        frontend.regIntfcObj(TechGetMosfetParams) 
        frontend.regIntfcObj(TechGetParam) 
    
       
              
