
import Plex
import sys
import exceptions

import indenter
prbuff = indenter.prbuff
prscan = indenter.prscan

import skintfc

# scaling factors:
#
#  Y - 10e24
#  Z - 10e21
#  E - 10e18
#  P - 10e15
#  T - 10e12
#  G - 10e9
#  M - 10e6
#  k/K - 10e3
#  % - 0.01
#  m - 10e-3
#  u - 10e-6
#  n - 10e-9
#  p - 10e-12
#  f - 10e-15
#  a - 10e-18
#  z - 10e-21
#  y - 10e-24
#

g_indentspaces = 4

class LexerException(Exception):
    pass

class Token:
# Token types
    TOK_NONE     = 0
    TOK_QUEST    = 1
    TOK_PLUS     = 2
    TOK_MINUS    = 3
    TOK_TIMES    = 4
    TOK_DIV      = 5
    TOK_LPAR     = 6
    TOK_RPAR     = 7
    TOK_EQUAL    = 8
    TOK_REDIR    = 9      # ~>

    TOK_OR       = 10
    TOK_AND      = 11
    TOK_EQ       = 12
    TOK_NE       = 13
    TOK_LT       = 14
    TOK_GT       = 15
    TOK_LE       = 16
    TOK_GE       = 17
    TOK_NOT      = 18
    TOK_COMMA    = 19

    TOK_COLON    = 20
    TOK_DOT      = 21
    TOK_TRUE     = 22
    TOK_FALSE    = 23
    TOK_PROG     = 24
    TOK_LBRACKET = 25
    TOK_RBRACKET = 26

    TOK_IF       = 30
    TOK_THEN     = 31
    TOK_ELSE     = 32
    TOK_LET      = 33
    TOK_LIST     = 34
    TOK_WHEN     = 35
    TOK_FOR      = 36
    TOK_FOREACH  = 37
    TOK_CASE     = 38
    TOK_RETURN   = 39

    TOK_IDENT    = 40
    TOK_FNCIDENT = 41
    TOK_PCELLDEF = 42
    TOK_ICON     = 43
    TOK_DCON     = 44
    TOK_SCON     = 45
    TOK_PROCEDURE= 46
    TOK_PCCELLVIEW= 47
    TOK_CAR      = 48
    TOK_CDR      = 49
    
    TOK_CADR     = 50
    TOK_CDDR     = 51
    TOK_BQUOTE   = 52
    TOK_KEYWORD  = 53
    TOK_UMINUS   = 54
    TOK_PLUSPLUS = 55
    TOK_MINUSMINUS=56
    TOK_UNLESS   = 57
    TOK_DECLARE  = 58
    TOK_RODCREATEPATH = 59
    
    TOK_RODCREATERECT = 60
    TOK_RODALIGN = 61

    TOK_EOI      = 70
    TOK_BLKCMNT  = 71        # not passed to parser
    TOK_EOLCMNT  = 72        # not passed to parser

    toknames = [" ", "?", "+", "-", '*', "/", "(", ")", "=", "~>",
                "or", "and", "==", "!=", "<", ">", "<=", ">=", "!", ",",
                ":", "t", "nil", "prog", "->", "[", "]", " ", " ", " ",
                "if", "then", "else", "let", "list", "when", "for", "foreach", "case", "return",
                "ident", "fncident", "pcDefinePCell", "integer", "float", "string", "procedure", "pcCellView", "car", "cdr",
                "cadr", "cddr", "`", "?", "-", "++", "--", "unless", "declare", "rodCreatePath", 
                "rodCreateRect", "rodAlign", " ", " ", " ", " ", " ", " ", " ", " ", 
                "eoi", "blkcmnt", "eolcmnt"]

    def __init__(self, type, value=None, filename=None, line=0, col=0):
        self.type = type
        if value is None:
            self.value = None
        else:
            self.value = str(value)
        self.filename = filename
        self.line = line
        self.col = col
        self.cmntAfter = ""
        self.cmntBefore = ""

    def gettype(self):
        return self.type

    def getvalue(self):
        return self.value

    def getfilename(self):
        return self.filename

    def getline_no(self):
        return self.line

    def getcol_no(self):
        return self.col

    def __getitem__(self, i):
        if i == 0:
            return self.type
        elif i == 1:
            return self.value
        elif i == 2:
            return self.filename
        elif i == 3:
            return self.line
        elif i == 4:
            return self.col
        return None

    def appendcmntafter(self, cmnt):
        self.cmntAfter = self.cmntAfter + cmnt

    def appendcmntbefore(self, cmnt):
        self.cmntBefore = self.cmntBefore + cmnt

class SkLexer(Plex.Scanner):

    def clearcmntbuffer(self):
        self.cmntbuffer = ""

    def appendcmntbuffer(self, text):
        self.cmntbuffer = self.cmntbuffer + text

    def getcmntbuffer(self):
        return self.cmntbuffer

    def beg_blockcomment(self, text):
        self.appendcmntbuffer(text)
        self.begin('block_comment')
        return None

    def end_blockcomment(self, text):
        self.appendcmntbuffer(text)
        # print "endblockcomment: " + self.cmntbuffer
        self.begin('')
        return Token.TOK_BLKCMNT

    def beg_eolcomment(self, text):
        self.appendcmntbuffer(text)
        self.begin('eolline_comment')
        return None

    def end_eolcomment(self, text):
        self.appendcmntbuffer(text)
        # print "endeolcomment: " + self.cmntbuffer
        self.begin('')
        return Token.TOK_EOLCMNT

    def accum_blockcomment(self, text):
        self.appendcmntbuffer(text)
        return None

    def accum_eolcomment(self, text):
        self.appendcmntbuffer(text)
        return None

    letter = Plex.Range("AZaz")
    digit =  Plex.Range("09")
    underscore = Plex.Str("_")
    quest = Plex.Str("?")

    builtin = Plex.Str("pcDefinePCell")
    rodcreatepath = Plex.Str("rodCreatePath")
    rodcreaterect = Plex.Str("rodCreateRect")
    rodalign = Plex.Str("rodAlign")

    name = Plex.Rep1(letter | underscore) +  Plex.Opt(Plex.Rep(letter | digit | underscore | quest))
    number =  Plex.Rep1(digit) + Plex.Opt(Plex.Str(".")) + Plex.Opt(Plex.Rep(digit))
    space =  Plex.Any(" \t\n\r")
    keyword = Plex.Rep1(quest) + Plex.Rep1(letter | underscore) +  Plex.Opt(Plex.Rep(letter | digit | underscore))

    beg_comment = Plex.Str("/*")
    end_comment = Plex.Str("*/")
    line_comment = Plex.Str(";")

    logop_and = Plex.Str("&&")
    logop_or = Plex.Str("||")
    logop_ne = Plex.Str("!=")
    logop_eq = Plex.Str("==")
    logop_gt = Plex.Str(">")
    logop_lt = Plex.Str("<")
    logop_ge = Plex.Str(">=")
    logop_le = Plex.Str("<=")
    redir    = Plex.Str("~>")
    redir2   = Plex.Str("->")

    strlit   =  Plex.Str('"') +  Plex.Rep( Plex.AnyBut('"')) +  Plex.Str('"')

    lexicon =  Plex.Lexicon([
        (Plex.Str("if"),       Token.TOK_IF),
        (Plex.Str("then"),     Token.TOK_THEN),
        (Plex.Str("else"),     Token.TOK_ELSE),
        (Plex.Str("let"),      Token.TOK_LET),
        (Plex.Str("list"),     Token.TOK_LIST),
        (Plex.Str("when"),     Token.TOK_WHEN),
        (Plex.Str("unless"),   Token.TOK_UNLESS),
        (Plex.Str("for"),      Token.TOK_FOR),
        (Plex.Str("foreach"),  Token.TOK_FOREACH),
        (Plex.Str("case"),     Token.TOK_CASE),
        (Plex.Str("return"),   Token.TOK_RETURN),
        (Plex.Str("declare"),  Token.TOK_DECLARE),
        (Plex.Str("t"),        Token.TOK_TRUE),
        (Plex.Str("nil"),      Token.TOK_FALSE),
        (Plex.Str("procedure"),Token.TOK_PROCEDURE),
        (Plex.Str("pcCellView"),Token.TOK_PCCELLVIEW),
        (Plex.Str("car"),      Token.TOK_CAR),
        (Plex.Str("cdr"),      Token.TOK_CDR),
        (Plex.Str("cadr"),     Token.TOK_CADR),
        (Plex.Str("cddr"),     Token.TOK_CDDR),
        (Plex.Str("prog"),     Token.TOK_PROG),   # not really handling this now
        (builtin,              Token.TOK_PCELLDEF),
        (rodcreatepath,        Token.TOK_RODCREATEPATH),
        (rodcreaterect,        Token.TOK_RODCREATERECT),
        (rodalign,             Token.TOK_RODALIGN),
        (name,                 Token.TOK_IDENT),
        (keyword,              Token.TOK_KEYWORD),
        (number,               Token.TOK_DCON),
        (strlit,               Token.TOK_SCON),
        (logop_and,            Token.TOK_AND),
        (logop_or,             Token.TOK_OR),
        (logop_ne,             Token.TOK_NE),
        (logop_eq,             Token.TOK_EQ),
        (logop_gt,             Token.TOK_GT),
        (logop_lt,             Token.TOK_LT),
        (logop_ge,             Token.TOK_GE),
        (logop_le,             Token.TOK_LE),
        (Plex.Str("?"),        Token.TOK_QUEST),
        (Plex.Str("+"),        Token.TOK_PLUS),
        (Plex.Str("-"),        Token.TOK_MINUS),
        (Plex.Str("*"),        Token.TOK_TIMES),
        (Plex.Str("/"),        Token.TOK_DIV),
        (Plex.Str("("),        Token.TOK_LPAR),
        (Plex.Str(")"),        Token.TOK_RPAR),
        (Plex.Str("="),        Token.TOK_EQUAL),
        (Plex.Str(","),        Token.TOK_COMMA),
        (Plex.Str(":"),        Token.TOK_COLON),
        (Plex.Str("."),        Token.TOK_DOT),
        (Plex.Str("'"),        Token.TOK_BQUOTE),
        (Plex.Str("["),        Token.TOK_LBRACKET),
        (Plex.Str("]"),        Token.TOK_RBRACKET),
        (Plex.Str("++"),       Token.TOK_PLUSPLUS),
        (Plex.Str("--"),       Token.TOK_MINUSMINUS),
        (redir,                Token.TOK_REDIR),
        (redir2,               Token.TOK_REDIR),
        (beg_comment,          beg_blockcomment),
        (line_comment,         beg_eolcomment),
        (Plex.Eof,             Token.TOK_EOI),
        (space,                Plex.IGNORE),
        Plex.State('block_comment', [
            (end_comment,           end_blockcomment),
            (Plex.AnyChar,          accum_blockcomment)
        ]),
        Plex.State('eolline_comment', [
            (Plex.Str("\n"),        end_eolcomment),
            (Plex.AnyChar,          accum_eolcomment)
        ])
    ])

    def __init__(self, infile):
        Plex.Scanner.__init__(self, self.lexicon, infile, infile)
        self.cmntbuffer = ""

###############################################################################

def formatpythoncomment(cmntbuff, spaceprefix=""):
    lines = cmntbuff.splitlines(True)
    ret = ""
    for l in lines:
        tmp = l
        if tmp[0] == ';':
            tmp = '#' + tmp[1:]
        else:
            tmp = '#' + tmp
        tmp = tmp.replace("/*", "")
        tmp = tmp.replace("*/", "")
        tmp = spaceprefix + tmp
        ret = ret + tmp
    
    return ret

###############################################################################


class ExprAttr:
    TV_NONE    = 0
    TV_BOOLLIT = 1
    TV_INTLIT  = 2
    TV_REALLIT = 3
    TV_STRLIT  = 4
    TV_BOOL    = 5
    TV_INTEGER = 6
    TV_REAL    = 7
    TV_STRING  = 8
    TV_OBJECT  = 9

class SymbolAttr:
    QT_NONE    = 10
    QT_PARAM   = 11
    QT_LOCAL   = 12
    QT_GLOBAL  = 13
    QT_FUNC    = 14

class Symbol:

    def __init__(self, name, rettype=ExprAttr.TV_NONE, rewriter=None):
        self.name = name
        self.rettype = rettype
        self.rewriter = rewriter
        self.qualtype = SymbolAttr.QT_NONE
#    def __repr__(self):
#        return "<symbol '%s'>" % self.name
    def __str__(self):
        return self.name
    def getname(self):
        return self.name
    def setname(self, newname):
        self.name = newname
    def getrewriter(self):
        return self.rewriter
    
    def typevalue(self):
        return self.rettype
    def settypevalue(self, type):
        self.rettype = type
    def qualifier(self):
        return self.qualtype
    def setqualifier(self, type):
        self.qualtype = type

class SymbolTable:
    def __init__(self, filename=""):
        self.symbols = {}
    def lookup(self, name):
        sym = self.symbols.get(name)
        return sym
    def insertexternal(self, name, rettype=ExprAttr.TV_NONE, rewriter=None):
        sym = self.symbols[name] = Symbol(name, rettype, rewriter)
        sym.setqualifier(SymbolAttr.QT_FUNC)
        return sym
    def insert(self, name, rettype=ExprAttr.TV_NONE, rewriter=None):
        sym = self.symbols[name] = Symbol(name, rettype, rewriter)
        return sym
    def haskey(self, name):
        return self.symbols.has_key(name)
    def delete(self, name):
        del self.symbols[name]

# becomes global - should be nicer way to do this though
st = SymbolTable()

def typetochar(attr):
    if attr == ExprAttr.TV_NONE:
        return "none"
    elif attr == ExprAttr.TV_BOOLLIT:
        return "bool"
    elif attr == ExprAttr.TV_INTLIT:
        return "int literal"
    elif attr == ExprAttr.TV_REALLIT:
        return "real literal"
    elif attr == ExprAttr.TV_STRLIT:
        return "string literal"
    elif attr == ExprAttr.TV_BOOL:
        return "bool"
    elif attr == ExprAttr.TV_INTEGER:
        return "integer"
    elif attr == ExprAttr.TV_REAL:
        return "real"
    elif attr == ExprAttr.TV_STRING:
        return "string"
    elif attr == ExprAttr.TV_OBJECT:
        return "object"
    else:
        return "unset (shouldn't happen)"

def qualtochar(qual):
    if qual == SymbolAttr.QT_NONE:
        return "none"
    elif qual == SymbolAttr.QT_PARAM:
        return "param"
    elif qual == SymbolAttr.QT_LOCAL:
        return "local"
    elif qual == SymbolAttr.QT_GLOBAL:
        return "global"
    elif qual == SymbolAttr.QT_FUNC:
        return "func"
    else:
        return "unset (shouldn't happen)"   
    
def dumpst():
    width = 20
    print "\nsymbol table contents..."
    for name, sym in st.symbols.items():
        typename = typetochar(sym.typevalue())
        rem = width - len(name)
        cnt = 1
        while rem < 1:
            rem = width + cnt*4 - len(name)
            cnt = cnt + 1
        print "\t" + name + " "*rem + "type = [" + typename + "]"
    
def dumpst2():
    # sort the keys - 
    width = 20
    print "\nsymbol table contents..."
    names = st.symbols.keys()
    names.sort()
    for name in names:
        sym = st.lookup(name)
        typename = typetochar(sym.typevalue())
        qualname = qualtochar(sym.qualifier())
        rem = width - len(name)
        cnt = 1
        while rem < 1:
            rem = width + cnt*4 - len(name)
            cnt = cnt + 1
        print "\t" + name + " "*rem + "type/qual = [" + typename + ", " + qualname + "]" 
        
###############################################################################

CNTXT_NONE   = 1
CNTXT_GLOBAL = 2
CNTXT_PCELL  = 3
CNTXT_LET    = 4
CNTXT_STMT   = 5

class Context:
    contextnames = ["", "cntxt::none", "cntxt::global", "cntxt::pcell", "cntxt::let", "cntxt::stmt"]
    
    def __init__(self, type = CNTXT_NONE, compstmt = None):
        self.prestmt = None
        self.poststmt = None
        self.type = type
        self.compstmt = compstmt
    
    def pre(self, cs):
        if self.prestmt != None:
            cs.append(self.prestmt)
            self.prestmt = None
    
    def post(self, cs):
        if self.poststmt != None:
            cs.append(self.poststmt)
            self.poststmt = None
            
    def slotPre(self, stmt):
        self.prestmt = stmt
        
    def slotPost(self, stmt):
        self.poststmt = stmt
        
    def type(self):
        return self.type
    
    def name(self):
        return self.contextnames[self.type]
    

class ContextStack:
    def __init__(self):
        self.stack = []

    def reset(self):
        self.stack = []

    def push(self, sv):
        # print "********* entering " + sv.name()
        self.stack.append(sv)

    def pop(self):
        # print "********* leaving " + self.top().name()
        return self.stack.pop()

    def isEmpty(self):
        return len(self.stack) == 0

    def top(self):
        # print "toplevel = " + self.stack[len(self.stack) - 1].name()
        return self.stack[len(self.stack) - 1]
     
cs = ContextStack()

###############################################################################

class Expr:
    def __init__(self):
        self.cmntBefore = ""
        self.cmntAfter = ""
        self.showParens = False
    def __str__(self):
        return ""
    def strrep(self):
        return self.__str__()
    def emit(self, redir, indent=0, cntxt=CNTXT_NONE):
        # default behavior
        redir.buff(self.spaces(indent) + self.__str__())
    def pppython(self, redir, indent=0, cntxt=CNTXT_NONE):
        # default behavior
        prbuff.substr(self.__str__())
    def installparams(self):
        raise SkParserError, "expr::installparams internal error"
    def defineparamspecs(self, indent, lineno=-1):
        raise SkParserError, "expr::defineparamspecs internal error"
    def setupparams(self, indent, lineno=-1):
        raise SkParserError, "expr::setupparams internal error"
    def asdeclaration(self, indent, cntxt):
        pass
    def redirdeclaration(self, redir, indent, cntxt):
        pass
    def repparens(self, exprstr):
        if self.showParens == True:
            return "(" + exprstr + ")"
        else:
            return exprstr
    def spaces(self, indent):
        tmp = ""
        if indent > 0:
            tmp += " "*indent
        return tmp
    def declarator(self):
        return ""
    def initializerexpr(self):
        return None
    def typevalue(self):
        return ExprAttr.TV_NONE
    def qualifier(self):
        return SymbolAttr.QT_NONE
    def setqualifier(self, qual):
        raise SkParserError, "expr::setqualifier internal error"
    def is_lval(self):   # lvals are 'writable'
        return False
    def is_container(self):
        return False
    def num_items(self):
        return 1
    def get_item(self,indx):
        if indx > 0:
            raise SkParseError, "expr::get_item internal error" 
        else:
            return self       

class BinaryExpr(Expr):
    def __init__(self, op, expr1, expr2):
        Expr.__init__(self)
        self.op = op
        self.expr1 = expr1
        self.expr2 = expr2
    def __str__(self):
        lstr = self.expr1.__str__()
        rstr = self.expr2.__str__()
        tmp = lstr + " " + Token.toknames[self.op] + " " + rstr
        return self.repparens(tmp)
    def pppython(self, redir, indent=0, cntxt=CNTXT_NONE):
        if self.showParens == True:
            prbuff.substr("(")
        self.expr1.pppython(redir, indent, cntxt)
        prbuff.blank(1)
        prbuff.substr(Token.toknames[self.op])
        prbuff.blank(1)
        self.expr2.pppython(redir, indent, cntxt)
        if self.showParens == True:
            prbuff.substr(")")
    def typevalue(self):
        if self.expr1.typevalue() == ExprAttr.TV_NONE or \
            self.expr2.typevalue() == ExprAttr.TV_NONE:
            return ExprAttr.TV_NONE        
        elif self.op == Token.TOK_AND or \
            self.op == Token.TOK_OR or \
            self.op == Token.TOK_EQ or \
            self.op == Token.TOK_NE or \
            self.op == Token.TOK_GT or \
            self.op == Token.TOK_LT or \
            self.op == Token.TOK_GE or \
            self.op == Token.TOK_LE:
            return TV_BOOL
        elif self.op ==Token.TOK_PLUS or \
            self.op == Token.TOK_MINUS or \
            self.op == Token.TOK_DIV or \
            self.op == Token.TOK_TIMES:
            if self.expr1.typevalue() == \
                self.expr2.typevalue():
                return self.expr1.typevalue()
            else:
                return ExprAttr.TV_REAL
                           
        return ExprAttr.TV_NONE

class PointExpr(Expr):
    def __init__(self, expr1, expr2):
        Expr.__init__(self)
        self.expr1 = expr1
        self.expr2 = expr2
    def __str__(self):
        lstr = self.expr1.__str__()
        rstr = self.expr2.__str__()
        tmp = "Point(" + lstr + ", " + rstr + ")"
        return self.repparens(tmp)
    def pppython(self, redir, indent=0, cntxt=CNTXT_NONE):
        if self.showParens == True:
            prbuff.substr("(")
        prbuff.substr("Point(")
        self.expr1.pppython(redir, indent, cntxt)
        prbuff.substr(",")
        prbuff.blank(1)
        self.expr2.pppython(redir, indent, cntxt)
        prbuff.substr(")")
        if self.showParens == True:
            prbuff.substr(")")
    def typevalue(self):
        return ExprAttr.TV_OBJECT

class KeywordExpr(Expr):
    def __init__(self, var, expr):
        Expr.__init__(self)
        self.var = var
        self.expr = expr
    def __str__(self):
        rstr = self.expr.__str__()
        tmp = self.var + "=" + rstr
        return tmp
    def pppython(self, redir, indent=0, cntxt=CNTXT_NONE):
        prbuff.substr(var)
        prbuff.blank(0)
        prbuff.substr("=")
        prbuff.blank(0)
        self.expr.pppython(redir, indent, cntxt)
    def typevalue(self):
        return ExprAttr.TV_NONE
    
class RedirExpr(Expr):
    def __init__(self, expr1, expr2):
        Expr.__init__(self)
        self.expr1 = expr1
        self.expr2 = expr2
    def __str__(self):
        ret = ""
        sym = st.lookup("dbGetq")
        if sym == None:
            lstr = self.expr1.__str__()
            rstr = self.expr2.__str__()
            tmp = "dbGetq(" + lstr + ", " + rstr + ")"
            return self.repparens(tmp)
        else:
            rewriter = sym.getrewriter()
            ret = rewriter.strrep([self.expr1, self.expr2])    
        return ret
    def pppython(self, redir, indent=0, cntxt=CNTXT_NONE):
        ret = ""
        sym = st.lookup("dbGetq")
        if sym == None:
            prbuff.substr("dbGetq(")
            self.expr1.pppython(redir, indent, cntxt)
            prbuff.substr(",")
            prbuff.blank(1)
            self.expr2.pppython(redir, indent, cntxt)
            prbuff.substr(")")
        else:
            rewriter = sym.getrewriter()
            ret = rewriter.pp(redir, [self.expr1, self.expr2])

class UnaryExpr(Expr):
    def __init__(self, op, expr):
        Expr.__init__(self)
        self.op = op
        self.expr = expr
    def __str__(self):
        tmp = Token.toknames[self.op] + self.expr.__str__()
        return self.repparens(tmp)
    def typevalue(self):
        return self.expr.typevalue()
    def pppython(self, redir, indent=0, cntxt=CNTXT_NONE):
        if self.showParens == True:
            prbuff.substr("(")
        self.expr.pppython(redir, indent, cntxt)
        if self.showParens == True:
            prebuff.substr(")")
#
# for PostFixExpr and PreFixExpr - the expr must be an lvalue (i.e.,
# a VarExpr).
#
class PostFixExpr(Expr):
    def __init__(self, op, expr):
        Expr.__init__(self)
        self.op = op
        self.expr = expr
        newop = Token.TOK_PLUS
        if self.op == Token.TOK_MINUSMINUS:
            newop = Token.TOK_MINUS
        stmt = makeIncrDecrStmt(expr, newop)
        cs.top().slotPost(stmt)
    def __str__(self):
        # ++/-- taken care of with slotting asgn stmt - just return sym
        return self.expr.__str__()
    def typevalue(self):
        return self.expr.typevalue()

class PreFixExpr(Expr):
    def __init__(self, op, expr):
        Expr.__init__(self)
        self.op = op
        self.expr = expr
        newop = Token.TOK_PLUS
        if self.op == Token.TOK_MINUSMINUS:
            newop = Token.TOK_MINUS
        stmt = makeIncrDecrStmt(expr, newop)
        cs.top().slotPre(stmt)
    def __str__(self):
        # ++/-- taken care of with slotting asgn stmt - just return sym
        return self.expr.__str__()
    def typevalue(self):
        return self.expr.typevalue()

class QuoteExpr(Expr):
    def __init__(self, expr):
        Expr.__init__(self)
        self.expr = expr
    def __str__(self):
        return "`" + self.expr.__str__()
    def typevalue(self):
        return self.expr.typevalue()

class ArrayIndexExpr(Expr):
    def __init__(self, expr, indexpr):
        Expr.__init__(self)
        self.expr = expr
        self.indexpr = indexpr
    def typevalue(self):
        return self.expr.typevalue()
    def __str__(self):
        return self.expr.__str__() + "[" + self.indexpr.__str__() + "]"
    def pppython(self, redir, indent=0, cntxt=CNTXT_NONE):
        self.expr.pppython(redir, indent, cntxt)
        prbuff.substr("[")
        self.indexpr.pppython(redir, indent, cntxt)
        prbuff.substr("]")

class LiteralExpr(Expr):
    def __init__(self, op, val):
        Expr.__init__(self)
        if op == Token.TOK_ICON:
            self.tv = ExprAttr.TV_INTEGER
        elif op == Token.TOK_DCON:
            self.tv = ExprAttr.TV_REAL
        elif op == Token.TOK_SCON:
            self.tv = ExprAttr.TV_STRING
        elif op == Token.TOK_PCCELLVIEW:
            self.tv = ExprAttr.TV_OBJECT
        else:
            raise SkParserError, "LiteralExp::__init__ internal error"
        self.op = op   # type of value
        self.val = val
    def __str__(self):
        return self.val
    def typevalue(self):
        return self.tv

class BoolExpr(Expr):
    def __init__(self, b):
        self.b = b
        self.tv = ExprAttr.TV_BOOL
    def __str__(self):
        if self.b == True:
            return "True"
        else:
            return "False"
    def typevalue(self):
        return self.tv

class VarExpr(Expr):
    # the sym(symbol) will always be defined within the symbol table.
    def __init__(self, sym):
        Expr.__init__(self)
        self.sym = sym
    def __str__(self):
        if self.sym.qualifier() == SymbolAttr.QT_PARAM:
            return "self." + self.sym.getname()
        else:
            return self.sym.getname()
    def typevalue(self):
        return self.sym.typevalue()
    def qualifier(self):
        return self.sym.qualifier()     
    def setqualifier(self, qual):
        return self.sym.setqualifier(qual)        
    def declarator(self):   
        return self.sym.getname()   
    def getlhs(self):
        return self.__str__()


class DeclareExpr(Expr):
    def __init__(self, expr, indexpr):
        Expr.__init__(self)
        self.expr = expr
        self.indexpr = indexpr
    def typevalue(self):
        return self.expr.typevalue()
    def __str__(self):
        return self.expr.__str__() + " = [None]*" + self.indexpr.__str__()
    def pppython(self, redir, indent=0, cntxt=CNTXT_NONE):
        self.expr.pppython(redir, indent, cntxt)
        prbuff.substr(" = ")
        prbuff.substr("[None]")
        prbuff.substr("*")
        self.indexpr.pppython(redir, indent, cntxt)

class NotExpr(Expr):
    def __init__(self, expr1):
        Expr.__init__(self)
        self.expr1 = expr1
    def __str__(self):
        return "not " + self.expr1
    
class CallExpr(Expr):
    def __init__(self, fncvar, args):
        Expr.__init__(self)
        self.fncvar = fncvar
        self.args = args
    def __str__(self):
        ret = ""
        sym = st.lookup(self.fncvar)
        if sym == None or sym.getrewriter() == "":
            ret = self.fncvar + "("
            for idx in range(0, len(self.args)):
                ret = ret + self.args[idx].__str__()
                if idx + 1 < len(self.args):
                    ret = ret + ", "
            ret = ret + ")"
        else:
            rewriter = sym.getrewriter()
            ret = rewriter.strrep(self.args)
        return ret
    def pppython(self, redir, indent=0, cntxt=CNTXT_NONE):
        ret = ""
        sym = st.lookup(self.fncvar)
        if sym == None or sym.getrewriter() == "":
            prbuff.substr(self.fncvar + "(")
            for idx in range(0, len(self.args)):
                self.args[idx].pppython(redir, indent, cntxt)
                if idx + 1 < len(self.args):
                    prbuff.substr(",")
                    prbuff.blank(1)
            prbuff.substr(")")
        else:
            rewriter = sym.getrewriter()
            rewriter.pp(redir, self.args)

    def typevalue(self):
        sym = st.lookup(self.fncvar)
        if sym == None:
            return ExprAttr.TV_NONE
        else:
            return sym.typevalue()

class RodCreatePathExpr(Expr):
    def __init__(self, fncvar, args):
        Expr.__init__(self)
        self.fncvar = fncvar
        self.args = args
    def __str__(self):
        # need to intercept the subRect parameter and rewrite
        # the embedded lists in a format suitable for calling
        # the rodCreatePath method
        ret = "self.rodCreatePath("
        for idx in range(0, len(self.args)):
            kwdval = self.args[idx]
            if isinstance(kwdval, KeywordExpr):
                var = kwdval.var
                if var == "subRect":
                    ret = ret + "subRect=["
                    for rectlist in kwdval.expr:
                        ret = ret + "["
                        for parset in rectlist:
                            ret = ret + "("
                            if isinstance(parset, KeywordExpr):
                                namedpar = parset.var
                                namedexpr = parset.expr
                                ret = ret + "\"" + namedpar + "\", "
                                if isinstance(namedexpr, LiteralExpr) and namedexpr.typevalue() == ExprAttr.TV_STRING:
                                    ret = ret + namedexpr.__str__()
                                else:
                                    ret = ret + "eval(\"" + namedexpr.__str__() + "\")"
                            ret = ret + "),"
                        ret = ret[:-1] + "],"    # strip extra comma
                    ret = ret[:-1] + "]"         # strip extra comma
                    if idx + 1 < len(self.args):
                        ret = ret + ", "
                elif var == "cvId":
                    # remove unnecessary clutter...
                    pass
                else:
                    ret = ret + kwdval.__str__()
                    if idx + 1 < len(self.args):
                        ret = ret + ", "
                    
        ret = ret + ")"
        return ret
    
    def pppython(self, redir, indent=0, cntxt=CNTXT_NONE):
        prbuff.substr(self.fncvar + "(")
        for idx in range(0, len(self.args)):
            self.args[idx].pppython(redir, indent, cntxt)
            if idx + 1 < len(self.args):
                prbuff.substr(",")
                prbuff.blank(1)
        prbuff.substr(")")

    def typevalue(self):
        return ExprAttr.TV_OBJECT

class RodCreateRectExpr(Expr):
    def __init__(self, fncvar, args):
        Expr.__init__(self)
        self.fncvar = fncvar
        self.args = args
    def __str__(self):
        # don't need to do anything for this one.
        ret = "self.rodCreateRect("
        for idx in range(0, len(self.args)):
            ret = ret + self.args[idx].__str__()
            if idx + 1 < len(self.args):
                ret = ret + ", "
        ret = ret + ")"
        return ret
    def pppython(self, redir, indent=0, cntxt=CNTXT_NONE):
        prbuff.substr(self.fncvar + "(")
        for idx in range(0, len(self.args)):
            self.args[idx].pppython(redir, indent, cntxt)
            if idx + 1 < len(self.args):
                prbuff.substr(",")
                prbuff.blank(1)
        prbuff.substr(")")

    def typevalue(self):
        return ExprAttr.TV_OBJECT
        
class CADRExpr(Expr):
    def __init__(self, op, expr):
        Expr.__init__(self)
        self.op = op
        self.expr = expr       
    def cadr_op_to_str(self):
        if self.op == Token.TOK_CAR:
            return "[0]"
        elif self.op == Token.TOK_CDR:
            return "[:1]"
        elif self.op == Token.TOK_CADR:
            return "[1]"
        elif self.op == Token.TOK_CDDR:
            return "[:2]"
        else:
            raise SkParser, "internal error in c[ad]*r"       
    def __str__(self):
        return self.repparens(self.expr.__str__() + self.cadr_op_to_str())
    def pppython(self, redir, indent=0, cntxt=CNTXT_NONE):
        prbuff.substr(self.expr.__str__())
        prbuff.blank(0)
        prbuff.substr(self.cadr_op_to_str)
    def typevalue(self):
        return ExprAttr.TV_NONE

class ListExpr(Expr):
    def __init__(self, explicit=False):
        Expr.__init__(self)
        self.elist = []
        self.isexplicit = explicit
    def __getitem__(self, idx):
        return self.elist[idx]
    def __len__(self):
        return len(self.elist)
    def is_container(self):
        return True
    def num_items(self):
        return len(self.elist)
    def get_item(self, indx):
        return self.elist[indx]
    def append(self, exp):
        self.elist.append(exp)
    def __str__(self):
        tmp = ""
        if self.isexplicit == True:
            tmp = "["
        if len(self.elist) == 0:
            tmp = tmp + "]"
            return tmp
        for idx in range(0, len(self.elist)):
            tmp2 = self.elist[idx].__str__()
            tmp = tmp + tmp2
            if idx + 1 < len(self.elist):
                tmp = tmp + ", "

        if self.isexplicit == True:
            tmp = tmp + "]"
        return self.repparens(tmp)
    def pppython(self, redir, indent=0, cntxt=CNTXT_NONE):
        if self.showParens == True:
            prbuff.substr("(")
        if self.isexplicit == True:
            prbuff.substr("[")
        if len(self.elist) == 0:
            prbuff.substr("]")
            if self.showParens == True:
                prbuff.substr(")")
            return
        for idx in range(0, len(self.elist)):
            self.elist[idx].pppython(redir, indent, cntxt)
            if idx + 1 < len(self.elist):
                prbuff.substr(",")
                prbuff.blank(1)

        if self.isexplicit == True:
            prbuff.substr("]")
        if self.showParens == True:
            prbuff.substr(")")
        return
    def typevalue(self):
        return ExprAttr.TV_OBJECT
    def declarator(self):
        ret = ""
        if len(self.elist) > 0:
            ret = self.elist[0].declarator()
        return ret
    def initializerexpr(self):
        ret = None
        if len(self.elist) > 2:
            ret = self.elist[2]
        elif len(self.elist) > 1:
            ret = self.elist[1]
        return ret
    def asdeclaration(self, indent, cntxt, lineno=-1):
        for idx in range(0, len(self.elist)):
            sp = self.spaces(indent)
            var = self.elist[idx].declarator()
            if self.elist[idx].initializerexpr() == None:
                str = sp + "# " + var.__str__()
            else:
                str = sp + var.__str__() + " = " + \
                    self.elist[idx].initializerexpr().__str__()
            print str        
    def redirdeclaration(self, redir, indent, cntxt, lineno=-1):
        for idx in range(0, len(self.elist)):
            sp = self.spaces(indent)
            var = self.elist[idx].declarator()
            if self.elist[idx].initializerexpr() == None:
                str = sp + "# " + var.__str__()
            else:
                str = sp + var.__str__() + " = " + \
                    self.elist[idx].initializerexpr().__str__()
            redir.buff(str, lineno)        
    def ppythondeclaration(self, redir, indent=0, cntxt=CNTXT_NONE):
        for idx in range(0, len(self.elist)):
            # prbuff.substr(self.spaces(indent))
            var = self.elist[idx].declarator()
            if self.elist[idx].initializerexpr() == None:
                prbuff.substr("# ");
                prbuff.substr(var.__str__())
            else:
                prbuff.substr(var.__str__())
                prbuff.blank(1)
                prbuff.substr("=")
                prubff.blank(1)
                self.elist[idx].pppython(redir, indent, cntxt)
            if idx < len(self.elist) - 1:
                prbuff.newline()
                     
    def defineparamspecs(self, redir, indent, lineno=-1):
        for idx in range(0, len(self.elist)):
            str = self.spaces(indent) + "specs(\'"
            var = self.elist[idx].declarator()
            str += var.__str__() + "\', "
            init = "0.0"
            if self.elist[idx].initializerexpr() != None:
                init = self.elist[idx].initializerexpr().__str__()
            str += init + ")"
            redir.write(str + "\n")
            
    def installparams(self):
        # set parameter up in the symbol table
        for idx in range(0, len(self.elist)):
            var = self.elist[idx].declarator()
            sym = st.lookup(var)
            init = self.elist[idx].initializerexpr()
            typeval = ExprAttr.TV_NONE
            if init != None:
                typeval = init.typevalue()
            if sym == None:
                sym = st.insert(var, typeval)
            else:
                sym.settypevalue(typeval)
            sym.setqualifier(SymbolAttr.QT_PARAM)

    def setupparams(self, redir, indent, lineno=-1):
        for idx in range(0, len(self.elist)):
            str = self.spaces(indent) + "self."
            var = self.elist[idx].declarator()
            str += var.__str__() + " = params[\'" + var.__str__() + "\']"
            redir.buff(str, lineno)
            
class Stmt:
    def __init__(self, lineno=-1):
        self.cmntBefore = ""
        self.cmntAfter = ""
        self.sklineno = lineno
    def __str__(self):
        ret = ""
        return ret
    def emit(self, redir, indent=0, cntxt=CNTXT_NONE):
        redir.buff("")
    def pppython(self, redir, indent=0, cntxt=CNTXT_NONE):
        prbuff.substr("")
    def count(self):
        return 0
    def spaces(self, indent):
        str = ""
        if indent > 0:
            str += " "*indent
        return str

class StmtList(Stmt):
    def __init__(self, stmts):
        Stmt.__init__(self)
        self.stmts = stmts
    def __str__(self):
        ret = ""
        for stmt in self.stmts:
            ret += stmt.__str__() + "\n"
        return ret
    def __getitem__(self, idx):
        return self.elist[idx]

class CompStmt(Stmt):
    def __init__(self):
        Stmt.__init__(self)
        self.stmts = []
    def __str__(self):
        ret = ""
        for stmt in self.stmts:
            ret += stmt.__str__() + "\n"
        return ret
    def __getitem__(self, idx):
        return self.stmts[idx]
    def count(self):
        return len(self.stmts)
    def append(self, stmt):
        self.stmts.append(stmt)
    def emit(self, redir, indent=0, cntxt=CNTXT_NONE):
        for idx in range(0, len(self.stmts)):
             self.stmts[idx].emit(redir, indent, cntxt)
    def pppython(self, redir, indent=0, cntxt=CNTXT_NONE):
        for idx in range(0, len(self.stmts)):
            if cntxt != CNTXT_GLOBAL:
                prbuff.newline()
            self.stmts[idx].pppython(redir, indent, cntxt)

class ExprStmt(Stmt):
    def __init__(self, expr, lineno):
        Stmt.__init__(self, lineno)
        self.expr = expr
    def __str__(self):
        return self.expr.__str__()
    def emit(self, redir, indent=0, cntxt=CNTXT_NONE):
        if self.expr.is_container() == False:
            redir.buff(self.spaces(indent) + self.__str__(), self.sklineno)
        else:
            raise SkParserError, "ExprStmt::emit declaration"
    def pppython(self, redir, indent=0, cntxt=CNTXT_NONE):
        if self.expr.is_container() == False:
            self.expr.pppython(redir, indent, cntxt)
        else:
            raise SkParserError, "ExprStmt::pppython declaration"

def mapSkillToPythonLoc(loc):
    loc = loc[1:-1]
    if loc == "upperLeft":
        return "Location.UPPER_LEFT"
    elif loc == "upperCenter":
        return "Location.UPPER_CENTER"
    elif loc == "upperRight":
        return "Location.UPPER_RIGHT"
    elif loc == "centerLeft":
        return "Location.CENTER_LEFT"
    elif loc == "centerCenter":
        return "Location.CENTER_CENTER"
    elif loc == "centerRight":
        return "Location.CENTER_RIGHT"
    elif loc == "lowerLeft":
        return "Location.LOWER_LEFT"
    elif loc == "lowerCenter":
        return "Location.LOWER_CENTER"
    elif loc == "lowerRight":
        return "Location.LOWER_RIGHT"
    else:
        return loc

class RodAlignStmt(Stmt):
    def __init__(self, expr, lineno):
        Stmt.__init__(self, lineno)
        self.expr = expr
    def __str__(self):
        return self.expr.__str__()
    def emit(self, redir, indent=0, cntxt=CNTXT_NONE):
        # just need to collect all the appropriate arguments - then rewrite
        refObj = None
        refHandle = None
        alignObj = None
        alignHandle = None
        xsep = 0.0
        ysep = 0.0
        args = self.expr.args
        for arg in args:
            if isinstance(arg, KeywordExpr):
                namedpar = arg.var
                namedexpr = arg.expr
                if namedpar == "refObj":
                    refObj = namedexpr
                elif namedpar == "refHandle":
                    refHandle = namedexpr
                elif namedpar == "alignObj":
                    alignObj = namedexpr
                elif namedpar == "alignHandle":
                    alignHandle = namedexpr
                elif namedpar == "xSep":
                    xsep = namedexpr
                elif namedpar == "ySep":
                    ysep = namedexpr
                
        if refObj == None or refHandle == None or alignObj == None or alignHandle == None:
            raise SkParserError, "RodAlignStmt:  insufficient parameters"
        #
        # now just emit the statements for alignment
        stmt1 = "_p1 = " + refObj.__str__()
        stmt1 = stmt1 + ".getBBox().getHandlePoint("
        stmt1 = stmt1 + mapSkillToPythonLoc(refHandle.__str__()) + ")"
        stmt2 = "_p1 = _p1 + Point(" + xsep.__str__() + ", " + ysep.__str__() + ")"
        stmt3 = "self.alignLocationToPoint(" + alignObj.__str__() + ", " + \
              mapSkillToPythonLoc(alignHandle.__str__()) + ", _p1)"
        redir.buff(self.spaces(indent) + stmt1, self.sklineno)
        redir.buff(self.spaces(indent) + stmt2, self.sklineno)
        redir.buff(self.spaces(indent) + stmt3, self.sklineno)
            
    def pppython(self, redir, indent=0, cntxt=CNTXT_NONE):
        if self.expr.is_container() == False:
            self.expr.pppython(redir, indent, cntxt)
        else:
            raise SkParserError, "RodAlignStmt::pppython declaration"

class LetStmt(Stmt):
    def __init__(self, decls, stmts, lineno):
        Stmt.__init__(self, lineno)
        self.decls = decls
        self.stmts = stmts
    def __str__(self):
        ret = "let(\n"
        ret += decls.__str__()
        for stmt in self.stmts:
            ret += stmt.__str__() + "; "
        return ret + ")"
    def emit(self, redir, indent=0, cntxt=CNTXT_NONE):
        # let does not adjust indent
        self.decls.redirdeclaration(redir, indent, CNTXT_LET, self.sklineno)
        for stmt in self.stmts:
            stmt.emit(redir, indent, CNTXT_LET)
    def pppython(self, redir, indent=0, cntxt=CNTXT_NONE):
        # let does not adjust indent
        #self.decls.ppythondeclaration(redir, indent, CNTXT_LET)
        #for stmt in self.stmts:
        #    stmt.pppython(redir, indent, CNTXT_LET)
        if cntxt == CNTXT_GLOBAL:
            prbuff.fullreset()        
            prbuff.openblock(indent)
            prbuff.newline()
            self.decls.ppythondeclaration(redir, indent, CNTXT_LET)
            prbuff.closeblock()
            prbuff.fmtctrl(indenter.PR_EOF)
            prbuff.reset()
            prscan()
            for stmt in self.stmts:
                prbuff.fullreset()
                prbuff.openblock(indent)
                prbuff.newline()
                stmt.pppython(redir, indent, CNTXT_LET)
                prbuff.closeblock()
                prbuff.fmtctrl(indenter.PR_EOF)
                prbuff.reset()
                prscan()
        else:     
            prbuff.openblock(indent)
            prbuff.newline()
            self.decls.ppythondeclaration(redir, indent, CNTXT_LET)
            prbuff.closeblock()           
            for stmt in self.stmts:
                prbuff.newline()
                stmt.pppython(redir, indent, CNTXT_LET)

class AsgnStmt(Stmt):
    def __init__(self, var, expr, lineno):
        Stmt.__init__(self, lineno)
        self.var = var
        self.expr = expr
        self.var.rettype = self.expr.typevalue()
    def __str__(self):
        return self.var + " = " + self.exp.__str__()
    def emit(self, redir, indent=0, cntxt=CNTXT_NONE):
        if len(self.cmntBefore) > 0:
            redir.buff(formatpythoncomment(self.cmntBefore, self.spaces(indent)), self.sklineno)
        t1 = self.expr.__str__()
        str = self.spaces(indent) + self.var.__str__() + " = " + t1
        if len(self.cmntAfter) > 0:
            str = str + " " + formatpythoncomment(self.cmntAfter)
        redir.buff(str, self.sklineno)
    def pppython(self, redir, indent=0, cntxt=CNTXT_NONE):
        if len(self.cmntBefore) > 0:
            prbuff.newline()
            prbuff.substr(formatpythoncomment(self.cmntBefore, self.spaces(indent)))       
        prbuff.newline()
        prbuff.substr(self.var.__str__())
        prbuff.blank(1)
        prbuff.substr("=")
        prbuff.blank(1)
        self.expr.pppython(redir, indent, cntxt)
        if len(self.cmntAfter) > 0:
            prbuff.substr(" " + formatpythoncomment(self.cmntAfter))

# used to support incr/decr operators
def makeIncrDecrStmt(varexpr, op):
    litexpr = LiteralExpr(Token.TOK_ICON, "1")
    asgnexpr = BinaryExpr(op, varexpr, litexpr)
    return AsgnStmt(varexpr, asgnexpr, 0)
    
class DbAsgnStmt(Stmt):
    def __init__(self, var, attr, expr, lineno):
        Stmt.__init__(self, lineno)
        self.var = var
        self.attr = attr
        self.expr = expr
    def __str__(self):
        return self.var + "~>" + self.attr + " = " + self.expr.__str__()   
    def emit(self, redir, indent=0, cntxt=CNTXT_NONE):
        if len(self.cmntBefore) > 0:
            redir.buff(formatpythoncomment(self.cmntBefore, self.spaces(indent)))
        ret = ""
        t1 = self.expr.__str__()
        ret = self.spaces(indent) + "dbSetq(" + self.var + ", " + self.attr + ", " + t1 + ")"
        if len(self.cmntAfter) > 0:
            ret = ret + " " + formatpythoncomment(self.cmntAfter)
        redir.buff(ret, self.sklineno)               
    def pppython(self, redir, indent=0, cntxt=CNTXT_NONE):
        if len(self.cmntBefore) > 0:
            prbuff.newline()
            prbuff.substr(formatpythoncomment(self.cmntBefore, self.spaces(indent)))       
        prbuff.newline()
        sym = st.lookup("dbSetq")
        if sym == None:
            prbuff.substr("dbSetq(")
            prbuff.blank(0)
            prbuff.substr(self.var)
            prbuff.blank(0)
            prbuff.substr(",")
            prbuff.blank(1)
            prbuff.substr(self.attr + ",")
            prbuff.blank(1)
            self.expr.pppython(redir, indent, cntxt)
            prbuff.blank(0)
            prbuff.substr(")")
        else:
            lhsvar = st.lookup(self.var)            
            if lhsvar == None:
                st.insert(self.var, ExprAttr.TV_OBJECT, "")
                lhsvar = st.lookup(self.var)
            else:
                lhsvar.settypevalue(ExprAttr.TV_OBJECT)
            lhsattr = st.lookup(self.attr)            
            if lhsattr == None:
                st.insert(self.attr, ExprAttr.TV_OBJECT, "")
                lhsattr = st.lookup(self.attr)
            else:
                lhsattr.settypevalue(ExprAttr.TV_OBJECT)
                
            rewriter = sym.getrewriter()
            rewriter.pp(redir, [VarExpr(lhsvar), VarExpr(lhsattr), self.expr])

        if len(self.cmntAfter) > 0:
            prbuff.substr(" " + formatpythoncomment(self.cmntAfter))

class IfElseStmt(Stmt):
    def __init__(self, exp, tstmts, fstmts, iflineno, elselineno):
        Stmt.__init__(self, iflineno)
        self.exp = exp
        self.tstmts = tstmts
        self.fstmts = fstmts
        self.sklineno2 = elselineno
    def __str__(self):
        ret = "if " + self.exp.__str__() + ":\n    " + self.tstmts.__str__()
        if len(self.fstmts) > 0:
            ret += "else:\n" + self.fstmts.__str__()
        return ret
    def emit(self, redir, indent=0, cntxt=CNTXT_NONE):
        global g_indentspaces
        if len(self.cmntBefore) > 0:
            redir.buff(formatpythoncomment(self.cmntBefore, self.spaces(indent)))
        redir.buff(self.spaces(indent) + "if " + self.exp.__str__() + ":", self.sklineno)
        self.tstmts.emit(redir, indent + g_indentspaces, cntxt)
        if self.fstmts != None and self.fstmts.count() > 0:
            redir.buff(self.spaces(indent) + "else:", self.sklineno2)
        if self.fstmts != None and self.fstmts.count() > 0:
            self.fstmts.emit(redir, indent + g_indentspaces, cntxt)
    def pppython(self, redir, indent=0, cntxt=CNTXT_NONE):
        global g_indentspaces
        if len(self.cmntBefore) > 0:
            prbuff.newline()
            prbuff.substr(formatpythoncomment(self.cmntBefore, self.spaces(indent)))
        prbuff.newline()
        prbuff.substr("if " + self.exp.__str__() + ":")
        prbuff.openblock(indent + g_indentspaces)
        self.tstmts.pppython(redir, indent, cntxt)
        prbuff.closeblock()
        if self.fstmts != None and self.fstmts.count() > 0:
            prbuff.substr("else:")
        if self.fstmts != None and self.fstmts.count() > 0:
            prbuff.openblock(indent + g_indentspaces)
            self.fstmts.pppython(redir, indent, cntxt)
        prbuff.closeblock()
            
class WhenStmt(Stmt):
    def __init__(self, exp, stmts, lineno):
        Stmt.__init__(self, lineno)
        self.exp = exp
        self.stmts = stmts
    def __str__(self):
        return "if " + self.exp.__str__() + ":\n" + self.stmts.__str__()
    def emit(self, redir, indent=0, cntxt=CNTXT_NONE):
        global g_indentspaces
        redir.buff(self.spaces(indent) + "if " + self.exp.__str__() + ":", self.sklineno)
        self.stmts.emit(redir, indent + g_indentspaces, cntxt)
    def pppython(self, redir, indent=0, cntxt=CNTXT_NONE):
        global g_indentspaces
        prbuff.newline()
        prbuff.substr("if ")
        self.exp.pppython(redir, indent, cntxt)
        prbuff.substr(":")
        prbuff.openblock(indent + g_indentspaces)
        self.stmts.pppython(redir, indent, cntxt)
        prbuff.closeblock()
        
class UnlessStmt(Stmt):
    def __init__(self, exp, stmts, lineno):
        Stmt.__init__(self, lineno)
        self.exp = exp
        self.stmts = stmts
    def __str__(self):
        return "if not (" + self.exp.__str__() + " ):\n" + self.stmts.__str__()
    def emit(self, redir, indent=0, cntxt=CNTXT_NONE):
        global g_indentspaces
        redir.buff(self.spaces(indent) + "if not (" + self.exp.__str__() + " ):", self.sklineno)
        self.stmts.emit(redir, indent + g_indentspaces, cntxt)
    def pppython(self, redir, indent=0, cntxt=CNTXT_NONE):
        global g_indentspaces
        prbuff.newline()
        prbuff.substr("if not(")
        self.exp.pppython(redir, indent, cntxt)
        prbuff.substr("):")
        prbuff.openblock(indent + g_indentspaces)
        self.stmts.pppython(redir, indent, cntxt)
        prbuff.closeblock()
        
class ForStmt(Stmt):
    def __init__(self, asgnexpr, startexpr, stopexpr, stmts, lineno):
        Stmt.__init__(self, lineno)
        self.asgnexpr = asgnexpr
        self.startexpr = startexpr
        self.stopexpr = stopexpr
        self.stmts = stmts
    def __str__(self):
        ret = "for " + self.asgnexpr.__str__() + " in range(" + self.startexpr.__str__()
        ret += ", " + self.stopexpr.__str__() + " + 1):\n" + self.stmts.__str__()
        return ret
    def emit(self, redir, indent=0, cntxt=CNTXT_NONE):
        global g_indentspaces
        ret = self.spaces(indent) + "for " + self.asgnexpr.__str__() + " in range("
        redir.buff(ret + self.startexpr.__str__() + ", " + self.stopexpr.__str__() + " + 1):", self.sklineno)
        self.stmts.emit(redir, indent + g_indentspaces, cntxt)
    def pppython(self, redir, indent=0, cntxt=CNTXT_NONE):
        global g_indentspaces
        prbuff.substr("for " + self.asgnexpr.__str__() + " in range(")
        prbuff.substr(self.startexpr.__str__() + ", ")
        prbuff.substr(self.stopexpr.__str__() + " + 1):")
        prbuff.openblock(indent + g_indentspaces)
        self.stmts.pppython(redir, indent + g_indentspaces, cntxt)
        prbuff.closeblock()

class ForEachStmt(Stmt):
    def __init__(self, startexpr, stopexpr, stmts, lineno):
        Stmt.__init__(self, lineno)
        self.startexpr = startexpr
        self.stopexpr = stopexpr
        self.stmts = stmts
    def __str__(self):
        ret = "foreach " + self.startexpr.__str__()
        ret += " in " + self.stopexpr.__str__() + ":\n"
        ret += self.stmts.__str__()
        return ret
    def emit(self, redir, indent=0, cntxt=CNTXT_NONE):
        global g_indentspaces
        ret = self.spaces(indent) + "foreach " + self.startexpr.__str__()
        ret += " in " + self.stopexpr.__str__() + ":"
        redir.buff(ret, self.sklineno)
        self.stmts.emit(redir, indent + g_indentspaces, cntxt)
    def pppython(self, redir, indent=0, cntxt=CNTXT_NONE):
        prbuff.substr("foreach " + self.startexpr.__str__())
        prbuff.substr(" in " + self.stopexpr.__str__() + ":")
        prbuff.openblock(indent + g_indentspaces)
        self.stmts.pppython(redir, indent + g_indentspaces, cntxt)
        prbuff.closeblock()

class CaseStmt(Stmt):
    def __init__(self, cnstexpr, stmts, lineno):
        Stmt.__init__(self, lineno)
        self.cnstexpr = cnstexpr
        self.stmts = stmts        # this is a stmtlist
    def __str__(self):
        str = ""
        first = True
        for stmt in self.stmts:
            if first == True:
                str += "if " + self.cnstexpr.__str__() + " == " + stmt.expr.__str__() + ":\n"
                first = False
            else:
                str += "elif " + self.cnstexpr.__str__() + " == " + stmt.expr.__str__() + ":\n"
            str += stmt.stmts.__str__()
        return str
    def emit(self, redir, indent=0, cntxt=CNTXT_NONE):
        global g_indentspaces
        first = True
        for stmt in self.stmts:
            if first == True:
                str = self.spaces(indent) + "if "
                first = False
            else:
                str = self.spaces(indent) + "elif "
            str += self.cnstexpr.__str__() + " == " + stmt.expr.__str__() + ":"
            redir.buff(str, self.sklineno)
            stmt.emit(redir, indent + g_indentspaces, cntxt)
    def pppython(self, redir, indent=0, cntxt=CNTXT_NONE):
        global g_indentspaces
        first = True
        for stmt in self.stmts:
            if first == True:
                prbuff.substr("if ")
                first = False
            else:
                prbuff.newline()
                prbuff.substr("elif ")
            prbuff.substr(self.cnstexpr.__str__())
            prbuff.substr(" == ")
            prbuff.substr(stmt.expr.__str__() + ":")
            prubff.openblock(indent + g_indentspaces)
            prbuff.newline()
            stmt.pppython(redir, indent, cntxt)
            prbuff.closeblock()

class CaseStmtBranch(Stmt):
    def __init__(self, expr, stmts, lineno):
        Stmt.__init__(self, lineno)
        self.expr = expr
        self.stmts = stmts
    def __str__(self):
        return self.stmts.__str__()
    def emit(self, redir, indent=0, cntxt=CNTXT_NONE):
        for stmt in self.stmts:
            stmt.emit(redir, indent, cntxt)
    def pppython(self, redir, indent=0, cntxt=CNTXT_NONE):
        for stmt in self.stmts:
            stmt.pppython(redir, indent, cntxt)
            
class ProcedureDef(Stmt):
    def __init(self, hdr, code):
        Stmt.__init__(self)
        self.hdr = hdr
        self.code = code

class CellDefStmt(Stmt):
    def __init__(self, cbpc, cellname, views, params, code,
                 lineno, lineno2, lineno3, lineno4, lineno5):
        Stmt.__init__(self, lineno)
        self.cmntBefore = cbpc
        self.cellname = cellname.replace("\"","")
        self.views = views
        self.params = params
        self.code = code
        self.sklineno2 = lineno2
        self.sklineno3 = lineno3
        self.sklineno4 = lineno4
        self.sklineno5 = lineno5
        self.hasRodCreatePath = False
        self.hasRodCreateRect = False
        self.hasRodAlign = False
        
    def __str__(self):
        str = "pcDefinePCell(\n"
        str += "    " + self.views.__str__() + "\n"
        str += "    " + self.params.__str__() + "\n"
        str += "    " + self.code.__str__() + "\n"
        str += "\n   )"
        return str
    
    def emit(self, redir, indent=0, cntxt=CNTXT_NONE):
        global g_indentspaces
        if len(self.cmntBefore) > 0:
            redir.buff(formatpythoncomment(self.cmntBefore, ""))
        redir.buff("\n\nfrom cni.dlo import *\nimport math")
        if self.hasRodCreatePath:
            redir.buff("\ngrpcnt = 0")
        redir.buff("\n\nclass " + self.cellname + "(DloGen):", self.sklineno2)
        
        redir.buff("\n" + self.spaces(g_indentspaces) + \
            "@classmethod\n" + self.spaces(g_indentspaces) + \
            "def defineParamSpecs(cls, specs):", self.sklineno3)
        # define parameter and default values.
        # looks like specs('param1', defvalue)
        self.params.defineparamspecs(redir, 2*g_indentspaces, self.sklineno3)

        redir.buff("\n" + self.spaces(g_indentspaces) + \
            "def setupParams(self, params):", self.sklineno3)
        # process parameter values entered by the user.
        # looks like:   self.param1 = params['param1']
        self.params.setupparams(redir, 2*g_indentspaces, self.sklineno3)

        redir.buff("\n" + self.spaces(g_indentspaces) + \
            "def genLayout(self):", self.sklineno4)
        
        # just traverse the rest of the code
        self.code.emit(redir, 2*g_indentspaces, CNTXT_GLOBAL)
        if self.hasRodCreatePath:
            redir.buff(skintfc.moduleDefRodCreatePath)
        if self.hasRodCreateRect:
            redir.buff(skintfc.moduleDefRodCreateRect)
        if self.hasRodAlign:
            redir.buff(skintfc.moduleDefRodAlign)
        
    def pppython(self, redir, indent=0, cntxt=CNTXT_NONE):
        global g_indentspaces
        if len(self.cmntBefore) > 0:
            prbuff.newline()
            redir.write(formatpythoncomment(self.cmntBefore, self.spaces(indent)) + "\n")
        redir.write("\n\nfrom cni.dlo import *")
        redir.write("\n\nclass Layout(DloGen):\n")

        redir.buff("\n" + self.spaces(g_indentspaces) + \
            "@classmethod\n" + self.spaces(g_indentspaces) + \
            "def defineParamSpecs(cls, specs):")

        # define parameter and default values.
        # looks like specs('param1', defvalue)
        self.params.defineparamspecs(redir, 2*g_indentspaces)

        redir.write("\n" + self.spaces(g_indentspaces) + \
            "def setupParams(self, params):\n")
        # process parameter values entered by the user.
        # looks like:   self.param1 = params['param1']
        self.params.setupparams(redir, 2*g_indentspaces)

        redir.write("\n" + self.spaces(g_indentspaces) + \
            "def genLayout(self):")
        
        # just traverse the rest of the code
        self.code.pppython(redir, 2*g_indentspaces, CNTXT_GLOBAL)

#####################################################################################


class SkParserError(exceptions.Exception):
    pass

class SkLexerError(exceptions.Exception):
    pass

class SkParser:
    def __init__(self):
        self.tokens = None
        self.curTokType = None
        self.curTokValue = ""
        self.curTokName = ""
        self.curTokLine = 0
        self.curTokCol = 0
        self.cmntBefore = ""
        self.cmntAfter = ""
        self.context = 0
        self.trace = 0
        self.cmntBuffer = ""

        self.hasRodCreatePath = False
        self.hasRodCreateRect = False
        self.hasRodAlign = False
        
    def configure(self, margin, contindent, blockindent):
        global g_indentspaces
        g_indentspaces = blockindent
        prbuff.configure(margin=margin, contindent=contindent, blockindent=blockindent)   
        
    def toggleTrace(self):
        self.trace = not self.trace 
            
    def insert(self, name, rettype=ExprAttr.TV_NONE, rewriter=None):
        global st
        st.insert(name, rettype, rewriter)
        
    def insertexternal(self, name, rettype=ExprAttr.TV_NONE, rewriter=None):
        global st
        st.insertexternal(name, rettype, rewriter)

    def regIntfcObj(self, obj):
        global st
        if st.haskey(obj.who()):
            st.delete(obj.who())
        st.insertexternal(obj.who(), obj.type(), obj)
                
    def tok_next(self):
        tok = self.tokens[0]
        self.curTokType = tok.gettype()
        self.curTokValue = tok.getvalue()
        self.curTokName = Token.toknames[self.curTokType]
        self.curTokLine = tok.getline_no()
        self.curTokCol = tok.getcol_no()
        self.cmntBefore = tok.cmntBefore
        self.cmntAfter = tok.cmntAfter
        del self.tokens[0]
        while self.curTokType in (Plex.IGNORE, None):
            tok = self.tokens[0]
            self.curTokType = tok.gettype()
            self.curTokValue = tok.getvalue()
            self.curTokName = Token.toknames[self.curTokType]
            self.curTokLine = tok.getline_no()
            self.curTokCol = tok.getcol_no()
            self.cmntBefore = tok.cmntBefore
            self.cmntAfter = tok.cmntAfter
            del self.tokens[0]
        
        if self.trace != 0:
            print "next token:  line# %d (code = %d) \'%s\'" %(self.curTokLine, self.curTokType, self.curTokValue)
#        if len(self.cmntAfter) > 0 or len(self.cmntBefore) > 0:
#            print "cmnt before: " + self.cmntBefore + "cmnt after: " + self.cmntAfter

    def tok_peektype(self):
        if len(self.tokens) == 0:
            raise SkParserError, "SkParser::tok_peektype internal error."
        return self.tokens[0].gettype()

    def expect(self, exptok):
        if self.curTokType != exptok:
            raise SkParserError, "syntax error.  found %s." % self.curTokValue

    def parse_file(self, fname):
        stmtnodes = []

        if self.trace != 0:
            print "opening file: " + fname
            
        infile = open(fname, "r")
        scanner =  SkLexer(infile)

        # buffer up all the tokens - kind of odd, but for this application
        # doesn't really matter.
        result = []
        lasttoken = None
        lastline = 0
        lastcol = 0
        seenpcelldef = 0
        
        while 1:
            try:
                token = scanner.read()
            except:    
                raise SkLexerError
            
            (filename, line, col) = scanner.position()
            if token[0] is None:
                result.append(Token(Token.TOK_EOI, "", "", line, col))
                break

            # eol comments are append to the last token we found (including EOI)
            if token[0] == Token.TOK_EOLCMNT:
                # for eol comment, error check for previous tokens, then only
                # for the previous token on the same line as the comment.
                # else just buffer (assume block comment)
                if len(result) > 0:
                    tokobjprev = result[-1]
                    if lastline == line:
                        tokobjprev.appendcmntafter(scanner.getcmntbuffer())
                        scanner.clearcmntbuffer()
                continue
            if token[0] == Token.TOK_BLKCMNT:
                # we don't have to do anything with these tokens except remove them
                continue
            
            # print 'toktype: ' + str(token[0]) + '  tokname: ' + token[1]
            #
            # correct for function identifiers = can't do this in Plex scanner unfortunately
            # here, we take the last token off the list (TOK_IDENT) and replace it with a
            # new token of type TOK_FNCIDENT.  Then fall thru and stick the '(' on the
            # token list.
            if token[0] == Token.TOK_LPAR and \
                lasttoken != None and \
                lasttoken[0] == Token.TOK_IDENT:
                if lastline == line and lastcol + len(lasttoken[1]) == col:
                    tokobjprev = result[-1]
                    result.remove(tokobjprev)
                    newtokobj = Token(Token.TOK_FNCIDENT, lasttoken[1], \
                                      filename, lastline, lastcol)
                    result.append(newtokobj)

            newtokobj = Token(token[0], token[1], filename, line, col)
            result.append(newtokobj)
            #
            # dump comment of all types that have been buffered up, unto the pcDefinePCell
            # token as they tend to get lost (especially when the PCell definition is 
            # embedded within a procedure.
            #
            if token[0] == Token.TOK_PCELLDEF:
                seenpcelldef = 1
            
            if seenpcelldef == 1:
                newtokobj.appendcmntbefore(scanner.getcmntbuffer())
                scanner.clearcmntbuffer()
              
            lasttoken = token
            lastline = line
            lastcol = col

        # print "status: tokenization complete"
        
        self.tokens = result
        self.tok_next()
        stmtnodes = self.parse()
        if self.trace != 0:
            dumpst2()
            
        infile.close()
        
        return stmtnodes

    #
    # look for statements at the global level.    here, we only
    # accept assignment statements and pcDefinePCell()
    #
    def parse(self, cntxt=CNTXT_GLOBAL):
        cs.push(Context(CNTXT_GLOBAL))
        stmtnodes = []    # top-level stmts
        stmt = None

        while self.curTokType != Token.TOK_EOI:
            stmt = self.parse_stmt(cntxt)
            cs.top().pre(stmtnodes)
            stmtnodes.append(stmt)
            cs.top().post(stmtnodes)

        return stmtnodes

    def parse_stmt(self, cntxt):
        stmt = None
        if self.curTokType == Token.TOK_IF:
            skiflineno = 0 #  self.CurTokLine
            cb = self.cmntBefore
            tlist = None
            flist = None
            self.tok_next()
            if self.curTokType != Token.TOK_LPAR:
                raise SkParserError, "expecting \'(\' for if."
            self.tok_next()
            cndexpr = self.parse_root_expr()
            cndexpr.cmntAfter = self.cmntAfter
            if self.curTokType != Token.TOK_THEN:
                raise SkParserError, "expecting \'then\' for if."
            self.tok_next()
            tlist = self.parse_comp_stmt(cntxt, stopping=[Token.TOK_ELSE, Token.TOK_RPAR])
            flist = None
            skelselineno = -1
            if self.curTokType == Token.TOK_ELSE:
                self.tok_next()
                flist = self.parse_comp_stmt(cntxt, stopping=[Token.TOK_RPAR])
            if self.curTokType != Token.TOK_RPAR:
                raise SkParserError, "unmatched \')\' for if."
            stmt = IfElseStmt(cndexpr, tlist, flist, skiflineno, skelselineno)
            stmt.cmntBefore = cb
            stmt.cmntAfter = self.cmntAfter
            self.tok_next()
            if self.trace != 0:
                print "reduce if stmt (line = %d)"%skiflineno
            return stmt
        elif self.curTokType == Token.TOK_WHEN:
            tlist = []
            skwhenlineno = self.curTokLine
            self.tok_next()
            if self.curTokType != Token.TOK_LPAR:
                raise SkParserError, "expecting \'(\' for when."
            self.tok_next()
            cndexpr = self.parse_root_expr()
            tlist = self.parse_comp_stmt(cntxt, stopping=[Token.TOK_RPAR])
            if self.curTokType != Token.TOK_RPAR:
                raise SkParserError, "unmatched \')\' for when."
            stmt = WhenStmt(cndexpr, tlist, skwhenlineno)
            self.tok_next()
            if self.trace != 0:
                print "reduce when stmt (line = %d)"%skwhenlineno
            # print "reduce when-stmt"
            return stmt

        elif self.curTokType == Token.TOK_UNLESS:
            tlist = []
            skwhenlineno = self.curTokLine
            self.tok_next()
            if self.curTokType != Token.TOK_LPAR:
                raise SkParserError, "expecting \'(\' for unless."
            self.tok_next()
            cndexpr = self.parse_root_expr()
            tlist = self.parse_comp_stmt(cntxt, stopping=[Token.TOK_RPAR])
            if self.curTokType != Token.TOK_RPAR:
                raise SkParserError, "unmatched \')\' for unless."
            stmt = UnlessStmt(cndexpr, tlist, skwhenlineno)
            self.tok_next()
            if self.trace != 0:
                print "reduce unless stmt (line = %d)"%skwhenlineno
            # print "reduce when-stmt"
            return stmt
        
        elif self.curTokType in (Token.TOK_LET, Token.TOK_PROG):
            skletlineno = self.curTokLine
            if self.curTokType == Token.TOK_PROG:
                print "warning: \'prog\' interpreted as \'let\' on line %d."%(skletlineno)
            stmt = self.parse_let_stmt(cntxt)
#            self.tok_next()
#            if self.curTokType != Token.TOK_LPAR:
#                raise SkParserError, "expecting \'(\' for let."
#            self.tok_next()
#            stmts = self.parse_comp_stmt(CNTXT_LET, stopping=[Token.TOK_RPAR])
#            stmt = LetStmt(stmts)
#            if self.curTokType != Token.TOK_RPAR:
#                raise SkParserError, "unmatched \')\' for let."
#            self.tok_next()
            if self.trace != 0:
                print "reduce let stmt (line = %d)"%skletlineno
            return stmt
        elif self.curTokType == Token.TOK_CASE:
            skcaselineno = self.curTokLine
            slist = []
            self.tok_next()
            if self.curTokType != Token.TOK_LPAR:
                raise SkParserError, "expecting \'(\' for case."
            self.tok_next()
            caseexpr = self.parse_root_expr()
            while self.curTokType == Token.TOK_LPAR:
                stmt = self.parse_case_branch_stmt(cntxt)
                slist.append(stmt)
            if self.curTokType != Token.TOK_RPAR:
                raise SkParserError, "unmatched \')\' for case."
            stmt = CaseStmt(caseexpr, slist, skcaselineno)
            self.tok_next()
            if self.trace != 0:
                print "reduce case stmt (line = %d)"%skcaselineno
            return stmt
        elif self.curTokType == Token.TOK_FOR:
            skforlineno = self.curTokLine
            slist = []
            self.tok_next()
            if self.curTokType != Token.TOK_LPAR:
                raise SkParserError, "expecting \'(\' for for."
            self.tok_next()
            asgnexpr = self.parse_root_expr()
            initexpr = self.parse_root_expr()
            finalexpr = self.parse_root_expr()
            cstmt = self.parse_comp_stmt(cntxt, stopping=[Token.TOK_RPAR])
            if self.curTokType != Token.TOK_RPAR:
                raise SkParserError, "unmatched \')\' for for."
            stmt = ForStmt(asgnexpr, initexpr, finalexpr, cstmt, skforlineno)
            self.tok_next()
            if self.trace != 0:
                print "reduce for stmt (line = %d)"%skforlineno
            return stmt
        elif self.curTokType == Token.TOK_FOREACH:
            skforeachlineno = curTokLine
            slist = []
            self.tok_next()
            if self.curTokType != Token.TOK_LPAR:
                raise SkParserError, "expecting \'(\' for foreach."
            self.tok_next()
            initexpr = self.parse_root_expr()
            finalexpr = self.parse_root_expr()
            cstmt = self.parse_comp_stmt(cntxt, stopping=[Token.TOK_RPAR])
            if self.curTokType != Token.TOK_RPAR:
                raise SkParserError, "unmatched \')\' for foreach."
            stmt = ForEachStmt(initexpr, finalexpr, cstmt, skforeachlineno)
            self.tok_next()
            if self.trace != 0:
                print "reduce foreach stmt (line = %d)"%skforeachlineno
            return stmt
        elif self.curTokType == Token.TOK_PCELLDEF:
            skpcelllineno = self.curTokLine
            stmt = self.parse_cell_def(cntxt)
            if self.trace != 0:
                print "reduce pcDefineCell stmt (line = %d)"%skpcelllineno
            return stmt
        elif self.curTokType == Token.TOK_PROCEDURE:
            skproclineno = self.curTokLine
            stmt = self.parse_procedure(cntxt)
            if self.trace != 0:
                print "reduce procedure def stmt (line = %d)"%skproclineno
            return stmt
        elif self.curTokType == Token.TOK_IDENT:
            thisline = self.curTokLine
            if self.tok_peektype() == Token.TOK_EQUAL:
                stmt = self.parse_asgn_stmt(cntxt)
                if self.trace != 0:
                    print "reduce assignment stmt (line = %d)"%thisline        
            elif self.tok_peektype() == Token.TOK_LBRACKET:
                stmt = self.parse_array_asgn_stmt(cntxt)
                if self.trace != 0:
                    print "reduce assignment stmt (line = %d)"%thisline        
            elif self.tok_peektype() == Token.TOK_REDIR:           
                stmt = self.parse_db_asgn_stmt(cntxt)     
                if self.trace != 0:
                    print "reduce dbSetq stmt (line = %d)"%thisline
            else:
                stmt = self.parse_expr_stmt(cntxt)                
                if self.trace != 0:
                    print "reduce expression stmt (line = %d)"%thisline
            return stmt
        elif self.curTokType == Token.TOK_FNCIDENT:
            skcalllineno = self.curTokLine
            id = self.curTokValue
            self.tok_next()
            if self.curTokType != Token.TOK_LPAR:
                raise SkParserError, "unmatched \'(\' for call."
            self.tok_next()
            args = self.parse_list_expr(False)
            if self.curTokType != Token.TOK_RPAR:
                raise SkParserError, "unmatched \')\' for call."
            exp = CallExpr(id, args)
            stmt = ExprStmt(exp, skcalllineno)
            if self.trace != 0:
                print "reduce procedure call stmt (line = %d)"%skcalllineno
            self.tok_next()
            return stmt
        elif self.curTokType == Token.TOK_RODALIGN:
            skcalllineno = self.curTokLine
            id = self.curTokValue
            self.tok_next()
            if self.curTokType != Token.TOK_LPAR:
                raise SkParserError, "unmatched \'(\' for rodAlignStmt."
            self.tok_next()
            args = self.parse_list_expr(False)
            if self.curTokType != Token.TOK_RPAR:
                raise SkParserError, "unmatched \')\' for rodAlignStmt."
            exp = CallExpr(id, args)
            stmt = RodAlignStmt(exp, skcalllineno)
            if self.trace != 0:
                print "reduce procedure call stmt (line = %d)"%skcalllineno
            self.tok_next()
            self.hasRodAlign = True
            return stmt            
        elif self.curTokType == Token.TOK_DECLARE:
            skdecllineno = self.curTokLine
            self.tok_next()
            if self.curTokType != Token.TOK_LPAR:
                raise SkParserError, "unmatched \'(\' for declare."
            self.tok_next()
            if self.curTokType != Token.TOK_IDENT:
                raise SkParserError, "expecting array identifier for declare."
            id = self.curTokValue
            sym = st.lookup(id)
            if sym == None:
                sym = st.insert(id, ExprAttr.TV_NONE, "")
            varexp = VarExpr(sym)
            self.tok_next()
            if self.curTokType != Token.TOK_LBRACKET:
                raise SkParserError, "unmatched \'[\' for declare."
            self.tok_next()
            indexpr = self.parse_root_expr()
            if self.curTokType != Token.TOK_RBRACKET:
                raise SkParserError, "unmatched \']\' for declare."
            self.tok_next()
            if self.curTokType != Token.TOK_RPAR:
                raise SkParserError, "unmatched \')\' for declare."
            exp = DeclareExpr(varexp, indexpr)
            stmt = ExprStmt(exp, skdecllineno)
            if self.trace != 0:
                print "reduce procedure declare stmt (line = %d)"%skdecllineno
            self.tok_next()
            return stmt
        elif self.curTokType == Token.TOK_LPAR:
            skexprlineno = self.curTokLine
            self.tok_next()
            expr = self.parse_list_expr(False)
            if self.curTokType != Token.TOK_RPAR:
                raise SkParserError, "unmatched \')\' for call."
            stmt = ExprStmt(expr, skexprlineno)
            if self.trace != 0:
                print "reduce bracketed () stmt (line = %d)"%skexprlineno
            self.tok_next()
            return stmt
        else:
            thisline = self.curTokLine
            stmt = self.parse_expr_stmt(stmt)
            self.tok_next()
            if self.trace != 0:
                print "reduce expression stmt (line = %d)"%thisline
            return stmt

    def parse_comp_stmt(self, cntxt, stopping):
        cstmt = CompStmt()# parse a group of statements
        stmt = None
        while self.curTokType not in stopping:
            stmt = self.parse_stmt(cntxt)
            cs.top().pre(cstmt)
            cstmt.append(stmt)
            cs.top().post(cstmt)
        return cstmt

    def parse_let_stmt(self, stopping):
        skletlineno = self.curTokLine
        self.tok_next()
        if self.curTokType != Token.TOK_LPAR:
            raise SkParserError, "expecting \'(\' for let."
        self.tok_next()
        self.tok_next()
        cs.push(Context(CNTXT_LET))
        decls = self.parse_list_expr(False)  
        if self.curTokType != Token.TOK_RPAR:
            raise SkParserError, "unmatched \')\' for let."
        self.tok_next()
        stmts = self.parse_comp_stmt(CNTXT_LET, stopping=[Token.TOK_RPAR])
        stmt = LetStmt(decls, stmts, skletlineno)
        if self.curTokType != Token.TOK_RPAR:
            raise SkParserError, "unmatched \')\' for let."
        cs.pop()
        self.tok_next()

        return stmt

    def parse_case_branch_stmt(self, cntxt):
        # parse list of case branches - of format '(' intcon stmtlist ')'
        skcaselineno = self.curTokLine
        if self.curTokType != Token.TOK_LPAR:
            raise SkParserError, "unmatched \')\' for case branch."
        self.tok_next()
        expr = self.parse_root_expr()
        slist = self.parse_comp_stmt(cntxt, stopping=[Token.TOK_RPAR])
        if self.curTokType != Token.TOK_RPAR:
            raise SkParserError, "unmatched \')\' for case branch."

        stmt = CaseStmtBranch(expr, slist, skcaselineno)
        self.tok_next()
        return stmt

    def parse_expr_stmt(self, cntxt):
        skexprlineno = self.curTokLine
        exp = self.parse_root_expr()
        return ExprStmt(exp, skexprlineno)

    def parse_asgn_stmt(self, cntxt):
        skasgnlineno = self.curTokLine
        cb = self.cmntBefore
        var = self.curTokValue
        
        self.tok_next()
        if self.curTokType != Token.TOK_EQUAL:
            raise SkParserError, "parse:  expecting \'=\' in assignment."

        self.tok_next()
        expr = self.parse_root_expr()
        ca = self.cmntAfter

        sym = st.lookup(var)
        if sym == None:
            sym = st.insert(var, expr.typevalue(), "")
        sym.settypevalue(expr.typevalue())

        stmt = AsgnStmt(VarExpr(sym), expr, skasgnlineno)
        stmt.cmntAfter = ca
        stmt.cmntBefore = cb
        return stmt

    def parse_array_asgn_stmt(self, cntxt):
        skasgnlineno = self.curTokLine
        cb = self.cmntBefore
        var = self.curTokValue
        
        self.tok_next()
        if self.curTokType != Token.TOK_LBRACKET:
            raise SkParserError, "parse:  expecting \'[\' in array assignment."

        self.tok_next()
        indexpr = self.parse_root_expr()
        if self.curTokType != Token.TOK_RBRACKET:
            raise SkParserError, "parse:  expecting \']\' in array assignment."
        
        self.tok_next()
        if self.curTokType != Token.TOK_EQUAL:
            raise SkParserError, "parse:  expecting \'=\' in array assignment."
        
        self.tok_next()        
        expr = self.parse_root_expr()
        ca = self.cmntAfter

        sym = st.lookup(var)
        if sym == None:
            sym = st.insert(var, expr.typevalue(), "")
        sym.settypevalue(expr.typevalue())

        arrayexpr = ArrayIndexExpr(VarExpr(sym), indexpr)
        stmt = AsgnStmt(arrayexpr, expr, skasgnlineno)
        
        stmt.cmntAfter = ca
        stmt.cmntBefore = cb
        return stmt

    def parse_db_asgn_stmt(self, cntxt):
        skdbasgnlineno = self.curTokLine
        cb = self.cmntBefore
        var = self.curTokValue

        self.tok_next()
        if self.curTokType != Token.TOK_REDIR:
            raise SkParserError, "parse:  expecting \'~>\' in db assignment."

        self.tok_next()
        attr = self.curTokValue

        self.tok_next()
        if self.curTokType != Token.TOK_EQUAL:
            raise SkParserError, "parse:  expecting \'=\' in db assignment."

        self.tok_next()
        exp = self.parse_root_expr()
        ca = self.cmntAfter

        stmt = DbAsgnStmt(var, attr, exp, skdbasgnlineno)
        stmt.cmntAfter = ca
        stmt.cmntBefore = cb
        return stmt

#
# ignore the procedure definition - just return the statement.
#
    def parse_procedure(self, cntxt):
        self.tok_next()
        if self.curTokType != Token.TOK_LPAR:
            raise SkParserError, "expecting \'(\' for procedure."
        self.tok_next()

        cbv = self.cmntBefore
        views = self.parse_root_expr()
        views.cmntBefore = cbv
        views.cmntAfter = self.cmntAfter

        if self.curTokType != Token.TOK_PCELLDEF:
            raise SkParserError, "expecting pcDefinePCell for procedure."
            
        stmt = self.parse_cell_def(cntxt)
        
        if self.curTokType != Token.TOK_RPAR:
            raise SkParserError, "expecting \')\' for procedure."
        self.tok_next()
        
        return stmt
        
    def parse_cell_def(self, cntxt):
        # pcDefineCell(list(views) list(params) code)
        cs.push(Context(CNTXT_PCELL))
        skpcelllineno = self.curTokLine
        cbpc = self.cmntBefore
        self.tok_next()
        if self.curTokType != Token.TOK_LPAR:
            raise SkParserError, "expecting \'(\' for pcDefinePCell."
        self.tok_next()

        cbv = self.cmntBefore
        skviewslineno = self.curTokLine
        views = self.parse_root_expr()
        cellname = "unset"
        if views.is_container() == True:
            cellname = views.get_item(1).__str__()
        views.cmntBefore = cbv
        views.cmntAfter = self.cmntAfter

        cbp = self.cmntBefore
        skparamlineno = self.curTokLine
        params = self.parse_root_expr()
        params.installparams()
        params.cmntBefore = cbp
        params.cmntAfter = self.cmntAfter

        cbc = self.cmntBefore
        skcodelineno = self.curTokLine
        # code = self.parse_stmt(stopping=[Token.TOK_RPAR])
        code = self.parse_stmt(CNTXT_PCELL)
        code.cmntBefore = cbc
        code.cmntAfter = self.cmntAfter

        # parse the return value if it is there....
        cbr = self.cmntBefore
        skretvallineno = self.curTokLine
        retval = None
        if self.curTokType != Token.TOK_RPAR:
            retval = self.parse_root_expr()
            retval.cmntBefore = cbr
            retval.cmntAfter = self.cmntAfter

        stmt = CellDefStmt(cbpc, cellname, views, params, code, \
                skpcelllineno, skviewslineno, skparamlineno, skcodelineno, skretvallineno)
        if self.curTokType != Token.TOK_RPAR:
            raise SkParserError, "expecting \')\' for pcDefinePCell."
        self.tok_next()
        cs.pop()

        # set some attributes for any methods seen
        stmt.hasRodCreatePath = self.hasRodCreatePath
        stmt.hasRodCreateRect = self.hasRodCreateRect
        stmt.hasRodAlign = self.hasRodAlign
        
        return stmt

#
# the precedence is defined by the call order of the recursive descent
# parser.
#
    def parse_root_expr(self):
        exp = self.parse_and_expr()
        while self.curTokType == Token.TOK_OR:
            self.tok_next()
            left = exp
            right = self.parse_and_expr()
            exp = BinaryExpr(Token.TOK_OR, left, right)

        return exp

    def parse_and_expr(self):
        exp = self.parse_equal_expr()
        while self.curTokType == Token.TOK_AND:
            self.tok_next()
            left = exp
            right = self.parse_equal_expr()
            exp = BinaryExpr(Token.TOK_AND, left, right)

        return exp

    def parse_equal_expr(self):
        exp = self.parse_rel_expr()
        while self.curTokType in (Token.TOK_EQ, Token.TOK_NE):
            op = self.curTokType
            self.tok_next()
            left = exp
            right = self.parse_rel_expr()
            exp = BinaryExpr(op, left, right)

        return exp

    def parse_rel_expr(self):
        exp = self.parse_add_expr()
        while self.curTokType in (Token.TOK_LT, Token.TOK_GT, Token.TOK_LE, Token.TOK_GE, Token.TOK_COLON):
            op = self.curTokType
            self.tok_next()
            left = exp
            right = self.parse_add_expr()
            if op == Token.TOK_COLON:
                exp = PointExpr(left, right)
            else:
                exp = BinaryExpr(op, left, right)

        return exp

    def parse_add_expr(self):
        exp = self.parse_mult_expr()
        while self.curTokType in (Token.TOK_PLUS, Token.TOK_MINUS):
            op = self.curTokType
            self.tok_next()
            left = exp
            right = self.parse_mult_expr()
            exp = BinaryExpr(op, left, right)

        return exp

    def parse_mult_expr(self):
        exp = self.parse_primary_expr()
        while self.curTokType in (Token.TOK_TIMES, Token.TOK_DIV, Token.TOK_REDIR):
            op = self.curTokType
            self.tok_next()
            left = exp
            right = self.parse_primary_expr()
            if op == Token.TOK_REDIR:
                exp = RedirExpr(left, right)
            else:
                exp = BinaryExpr(op, left, right)

        return exp

    def parse_primary_expr(self):
        if self.curTokType == Token.TOK_LPAR:
            self.tok_next()
            exp = self.parse_list_expr(False)
            if self.curTokType != Token.TOK_RPAR:
                raise SkParserError, "unmatched \')\'"
            exp.showParens = True
        elif self.curTokType == Token.TOK_LIST:
            self.tok_next()
            if self.curTokType != Token.TOK_LPAR:
                raise SkParserError, "expecting \'(\' for list."
            self.tok_next()
            exp = self.parse_list_expr(True)
            if self.curTokType != Token.TOK_RPAR:
                raise SkParserError, "unmatched \')\' for list."
        elif self.curTokType in (Token.TOK_MINUS, Token.TOK_PLUS, \
                                 Token.TOK_PLUSPLUS, Token.TOK_MINUSMINUS):
            op = self.curTokType
            self.tok_next()
            prexp = self.parse_primary_expr()
            if op == Token.TOK_MINUS or op == Token.TOK_PLUS:
                exp = UnaryExpr(op, prexp)
            else:
                exp = PreFixExpr(op, prexp)
            return exp
        elif self.curTokType == Token.TOK_BQUOTE:
            self.tok_next()
            exp = QuoteExpr(self.parse_primary_expr())
            return exp
        elif self.curTokType == Token.TOK_KEYWORD:
            var = self.curTokValue[1:]
            self.tok_next()
            asgnexpr = self.parse_root_expr()
            exp = KeywordExpr(var, asgnexpr)
            return exp
        elif self.curTokType in (Token.TOK_SCON, Token.TOK_DCON, \
                                 Token.TOK_ICON, Token.TOK_PCCELLVIEW):
            exp = LiteralExpr(self.curTokType, self.curTokValue)
        elif self.curTokType in (Token.TOK_TRUE, Token.TOK_FALSE):
            exp = BoolExpr(self.curTokType == Token.TOK_TRUE)
        elif self.curTokType == Token.TOK_IDENT:
            id = self.curTokValue
            sym = st.lookup(id)
            if sym == None:
                sym = st.insert(id, ExprAttr.TV_NONE, "")
            exp = VarExpr(sym)
            if self.tok_peektype() == Token.TOK_LBRACKET:
                self.tok_next()  # skip to bracket
                self.tok_next()  # now skip over it...
                indexp = self.parse_root_expr()
                if self.curTokType != Token.TOK_RBRACKET:
                    raise SkParserError, "unmatched \']\' for array index."
                exp = ArrayIndexExpr(exp, indexp)
            elif self.tok_peektype() in (Token.TOK_PLUSPLUS, Token.TOK_MINUSMINUS):
                self.tok_next()
                exp = PostFixExpr(self.curTokType, exp)        
        elif self.curTokType == Token.TOK_FNCIDENT:
            id = self.curTokValue
            self.tok_next()
            if self.curTokType != Token.TOK_LPAR:
                raise SkParserError, "unmatched \'(\' for function call."
            self.tok_next()
            args = self.parse_list_expr(False)
            if self.curTokType != Token.TOK_RPAR:
                raise SkParserError, "unmatched \')\' for function call."
            exp = CallExpr(id, args)
        elif self.curTokType == Token.TOK_RODCREATEPATH:
            id = self.curTokValue
            self.tok_next()
            if self.curTokType != Token.TOK_LPAR:
                raise SkParserError, "unmatched \'(\' for rodCreatePath call."
            self.tok_next()
            args = self.parse_list_expr(False)
            if self.curTokType != Token.TOK_RPAR:
                raise SkParserError, "unmatched \')\' for rodCreatePath call."
            exp = RodCreatePathExpr(id, args)     
            self.hasRodCreatePath = True
        elif self.curTokType == Token.TOK_RODCREATERECT:
            id = self.curTokValue
            self.tok_next()
            if self.curTokType != Token.TOK_LPAR:
                raise SkParserError, "unmatched \'(\' for rodCreateRect call."
            self.tok_next()
            args = self.parse_list_expr(False)
            if self.curTokType != Token.TOK_RPAR:
                raise SkParserError, "unmatched \')\' for rodCreateRect call."
            exp = RodCreateRectExpr(id, args)          
            self.hasRodCreateRect = True
        elif self.curTokType in (Token.TOK_CAR, Token.TOK_CDR, \
                                 Token.TOK_CADR, Token.TOK_CDDR):
            op = self.curTokType
            self.tok_next()
            if self.curTokType != Token.TOK_LPAR:
                raise SkParserError, "unmatched \'(\' for c[ad]*r."
            self.tok_next()
            args = self.parse_list_expr(False)
            if self.curTokType != Token.TOK_RPAR:
                raise SkParserError, "unmatched \')\' for c[ad]*r."
            exp = CADRExpr(op, args)
        elif self.curTokType == Token.TOK_NOT:
            self.tok_next()
            exp = NotExpr(self.parse_primary_expr())
            return exp
        elif self.curTokType == Token.TOK_EOI:
            msg = "unexpectedly reached end of input (line %d, col %d)" %(self.curTokLine, self.curTokCol)
            raise SkParserError, msg
        elif self.curTokType != Token.TOK_RPAR:
            msg = "unexpected input: \'%s\'(line %d, col %d)" %(self.curTokValue, self.curTokLine, self.curTokCol)
            raise SkParserError, msg
        else:
            exp = 0

        self.tok_next()
        return exp


    def parse_list_expr(self, explicit):
        # handle empty lists
        elist = ListExpr(explicit)
        if self.curTokType == Token.TOK_RPAR:
            return elist
        exp = self.parse_root_expr()
        elist.append(exp)
        if self.curTokType == Token.TOK_COMMA:
            self.tok_next()

        while self.curTokType != Token.TOK_RPAR:
            exp = self.parse_root_expr()
            elist.append(exp)
            if self.curTokType == Token.TOK_COMMA:
                self.tok_next()

        return elist

#if __name__ == "__main__":
#    pt = SkParser()
#    st = SymbolTable()
#    f1 = r"C:\Documents and Settings\dan\Desktop\appwave\samp1.il"
#    try:
#        pt.parse_file(f1)
#    except (SkParserError, msg):
#        print "Error during parsing input" + msg
#    except:
#        print "exception thrown"
#    else:
#        print "all done!"
